import re
import traceback
import typing
import warnings
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from docutils.parsers.rst import Directive
from lsprotocol.types import CompletionItem
from lsprotocol.types import DocumentLink
from lsprotocol.types import Location
from lsprotocol.types import MarkupContent
from lsprotocol.types import MarkupKind
from lsprotocol.types import Position
from lsprotocol.types import Range

from esbonio.lsp import CompletionContext
from esbonio.lsp import DefinitionContext
from esbonio.lsp import DocumentLinkContext
from esbonio.lsp import HoverContext
from esbonio.lsp import ImplementationContext
from esbonio.lsp import LanguageFeature
from esbonio.lsp import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer
from esbonio.lsp.util.inspect import get_object_location
from esbonio.lsp.util.patterns import DIRECTIVE
from esbonio.lsp.util.patterns import DIRECTIVE_OPTION

from .completions import render_directive_completion
from .completions import render_directive_option_completion

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore[assignment]


class DirectiveLanguageFeature:
    """Base class for directive language features."""

    def complete_arguments(
        self, context: CompletionContext, domain: str, name: str
    ) -> List[CompletionItem]:
        """Return a list of completion items representing valid targets for the given
        directive.

        Parameters
        ----------
        context:
           The completion context
        domain:
           The name of the domain the directive is a member of
        name:
           The name of the domain
        """
        return []

    def get_implementation(
        self, directive: str, domain: Optional[str]
    ) -> Optional[Type[Directive]]:
        """Return the implementation for the given directive name."""
        return self.index_directives().get(directive, None)

    def index_directives(self) -> Dict[str, Type[Directive]]:
        """Return all known directives."""
        return dict()

    def suggest_directives(
        self, context: CompletionContext
    ) -> Iterable[Tuple[str, Type[Directive]]]:
        """Suggest directives that may be used, given a completion context."""
        return self.index_directives().items()

    def suggest_options(
        self, context: CompletionContext, directive: str, domain: Optional[str]
    ) -> Iterable[str]:
        """Suggest options that may be used, given a completion context."""

        impl = self.get_implementation(directive, domain)
        if impl is None:
            return []

        option_spec = getattr(impl, "option_spec", {}) or {}
        return option_spec.keys()

    def resolve_argument_link(
        self,
        context: DocumentLinkContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> Tuple[Optional[str], Optional[str]]:
        """Resolve a document link request for the given argument.

        Parameters
        ----------
        context
           The context of the document link request.

        directive
           The name of the directive the argument is associated with.

        domain
           The name of the domain the directive belongs to, if applicable.

        argument
           The argument to resolve the link for.
        """
        return None, None

    def find_argument_definitions(
        self,
        context: DefinitionContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> List[Location]:
        """Return a list of locations representing definitions of the given argument.

        Parameters
        ----------
        context
           The context of the definition request.

        directive
           The name of the directive the argument is associated with.

        domain
           The name of the domain the directive belongs to, if applicable.

        argument
           The argument to find the definition of.
        """
        return []


class ArgumentCompletion(Protocol):
    """A completion provider for directive arguments.

    .. deprecated:: 0.14.2

       This will be removed in ``v1.0``, use subclasses of
       :class:`~esbonio.lsp.directives.DirectiveLanguageFeature` instead.
    """

    def complete_arguments(
        self, context: CompletionContext, domain: str, name: str
    ) -> List[CompletionItem]:
        """Return a list of completion items representing valid targets for the given
        directive.

        Parameters
        ----------
        context:
           The completion context
        domain:
           The name of the domain the directive is a member of
        name:
           The name of the domain
        """


class ArgumentDefinition(Protocol):
    """A definition provider for directive arguments.

    .. deprecated:: 0.14.2

       This will be removed in ``v1.0``, use subclasses of
       :class:`~esbonio.lsp.directives.DirectiveLanguageFeature` instead.
    """

    def find_definitions(
        self,
        context: DefinitionContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> List[Location]:
        """Return a list of locations representing definitions of the given argument.

        Parameters
        ----------
        context:
           The context of the definition request.
        directive:
           The name of the directive the argument is associated with.
        domain:
           The name of the domain the directive belongs to, if applicable.
        argument:
           The argument to find the definition of.
        """


class ArgumentLink(Protocol):
    """A document link resolver for directive arguments.

    .. deprecated:: 0.14.2

       This will be removed in ``v1.0``, use subclasses of
       :class:`~esbonio.lsp.directives.DirectiveLanguageFeature` instead.
    """

    def resolve_link(
        self,
        context: DocumentLinkContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> Tuple[Optional[str], Optional[str]]:
        """Resolve a document link request for the given argument.

        Parameters
        ----------
        context:
           The context of the document link request.
        directive:
           The name of the directive the argument is associated with.
        domain:
           The name of the domain the directive belongs to, if applicable.
        argument:
           The argument to resolve the link for.
        """


class Directives(LanguageFeature):
    """Directive support for the language server."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._documentation: Dict[str, Dict[str, str]] = {}
        """Cache for documentation."""

        self._features: Dict[str, DirectiveLanguageFeature] = {}
        """The collection of registered features."""

    def add_feature(self, feature: DirectiveLanguageFeature):
        """Register a directive language feature.

        Parameters
        ----------
        feature
           The directive language feature
        """
        key = f"{feature.__module__}.{feature.__class__.__name__}"

        # Create an unique key for this instance.
        if key in self._features:
            key += f".{len([k for k in self._features.keys() if k.startswith(key)])}"

        self._features[key] = feature

    def add_argument_completion_provider(self, provider: ArgumentCompletion) -> None:
        """Register an :class:`~esbonio.lsp.directives.ArgumentCompletion` provider.

        .. deprecated:: 0.14.2

           This will be removed in ``v1.0``, use
           :meth:`~esbonio.lsp.directives.Directives.add_feature` with a
           :class:`~esbonio.lsp.directives.DirectiveLanguageFeature` subclass instead.

        Parameters
        ----------
        provider:
           The provider to register.
        """
        warnings.warn(
            "ArgumentCompletion providers are deprecated in favour of "
            "DirectiveLanguageFeatures, this method will be removed in v1.0",
            DeprecationWarning,
            stacklevel=2,
        )

        name = provider.__class__.__name__
        key = f"{provider.__module__}.{name}.completion"

        # Automatically derive the feature definition from the provider.
        feature = type(
            f"{name}CompletionProvider",
            (DirectiveLanguageFeature,),
            {"complete_arguments": provider.complete_arguments},
        )()

        self._features[key] = feature

    def add_argument_definition_provider(self, provider: ArgumentDefinition) -> None:
        """Register an :class:`~esbonio.lsp.directives.ArgumentDefinition` provider.

        .. deprecated:: 0.14.2

           This will be removed in ``v1.0``, use
           :meth:`~esbonio.lsp.directives.Directives.add_feature` with a
           :class:`~esbonio.lsp.directives.DirectiveLanguageFeature` subclass instead.

        Parameters
        ----------
        provider:
           The provider to register.
        """
        warnings.warn(
            "ArgumentDefinition providers are deprecated in favour of "
            "DirectiveLanguageFeatures, this method will be removed in v1.0",
            DeprecationWarning,
            stacklevel=2,
        )

        name = provider.__class__.__name__
        key = f"{provider.__module__}.{name}.definitions"

        # Automatically derive the feature definition from the provider.
        feature = type(
            f"{name}DefinitionProvider",
            (DirectiveLanguageFeature,),
            {"find_argument_definitions": provider.find_definitions},
        )()

        self._features[key] = feature

    def add_argument_link_provider(self, provider: ArgumentLink) -> None:
        """Register an :class:`~esbonio.lsp.directives.ArgumentLink` provider.

        .. deprecated:: 0.14.2

           This will be removed in ``v1.0``, use
           :meth:`~esbonio.lsp.directives.Directives.add_feature` with a
           :class:`~esbonio.lsp.directives.DirectiveLanguageFeature` subclass instead.

        Parameters
        ----------
        provider:
           The provider to register.
        """
        warnings.warn(
            "ArgumentLink providers are deprecated in favour of "
            "DirectiveLanguageFeatures, this method will be removed in v1.0",
            DeprecationWarning,
            stacklevel=2,
        )

        name = provider.__class__.__name__
        key = f"{provider.__module__}.{name}.links"

        # Automatically derive the feature definition from the provider.
        feature = type(
            f"{name}LinkProvider",
            (DirectiveLanguageFeature,),
            {"resolve_argument_link": provider.resolve_link},
        )()

        self._features[key] = feature

    def add_documentation(self, documentation: Dict[str, Dict[str, Any]]) -> None:
        """Register directive documentation.

        ``documentation`` should be a dictionary with the following structure ::

           documentation = {
               "raw(docutils.parsers.rst.directives.misc.Raw)": {
                   "is_markdown": true,
                   "license": "https://...",
                   "source": "https://...",
                   "description": [
                       "# .. raw::",
                       "The raw directive is used for...",
                       ...
                    ]
                   "options": {
                       "file": "The file option allows...",
                       ...
                   }
               }
           }

        where the key has the form ``name(dotted_name)``. There are cases where a
        directive's implementation is not sufficient to uniquely identify it as
        multiple directives can be provided by a single class.

        This means the key has to be a combination of the ``name`` the user writes
        in a reStructuredText document and ``dotted_name`` is the fully qualified
        class name of the directive's implementation.

        .. note::

           If there is a clash with an existing key, the existing value will be
           overwritten with the new value.

        The values in this dictionary are themselves dictionaries with the following
        fields.

        ``description``
           A list of strings for the directive's main documentation.

        ``options``,
           A dictionary, with a field for the documentaton of each of the directive's
           options.

        ``is_markdown``
           A boolean flag used to indicate whether the ``description`` and ``options``
           are written in plain text or markdown.

        ``source``
           The url to the documentation's source

        ``license``
           The url to the documentation's license

        Parameters
        ----------
        documentation:
           The documentation to register.
        """

        for key, doc in documentation.items():
            description = doc.get("description", [])

            if not description:
                continue

            source = doc.get("source", "")
            if source:
                description.append(f"\n[Source]({source})")

            license = doc.get("license", "")
            if license:
                description.append(f"\n[License]({license})")

            doc["description"] = "\n".join(description)
            self._documentation[key] = doc

    def get_directives(self) -> Dict[str, Type[Directive]]:
        """Return a dictionary of all known directives."""

        directives = {}

        for name, feature in self._features.items():
            try:
                directives.update(feature.index_directives())
            except Exception:
                self.logger.error(
                    "Unable to index directives, error in feature '%s'\n%s",
                    name,
                    traceback.format_exc(),
                )

        return directives

    def get_implementation(
        self, directive: str, domain: Optional[str]
    ) -> Optional[Type[Directive]]:
        """Return the implementation of a directive given its name

        Parameters
        ----------
        directive
           The name of the directive.

        domain
           The domain of the directive, if applicable.
        """

        if domain:
            name = f"{domain}:{directive}"
        else:
            name = directive

        for feature_name, feature in self._features.items():
            try:
                impl = feature.get_implementation(directive, domain)
                if impl is not None:
                    return impl
            except Exception:
                self.logger.error(
                    "Unable to get implementation for '%s', error in feature: '%s'\n%s",
                    name,
                    feature_name,
                    traceback.format_exc(),
                )

        self.logger.debug(
            "Unable to get implementation for '%s', unknown directive", name
        )
        return None

    def suggest_directives(
        self, context: CompletionContext
    ) -> Iterable[Tuple[str, Type[Directive]]]:
        """Suggest directives that may be used, given a completion context.

        Parameters
        ----------
        context
           The CompletionContext.
        """

        for name, feature in self._features.items():
            try:
                yield from feature.suggest_directives(context)
            except Exception:
                self.logger.error(
                    "Unable to suggest directives, error in feature: '%s'\n%s",
                    name,
                    traceback.format_exc(),
                )

    def suggest_options(
        self, context: CompletionContext, directive: str, domain: Optional[str]
    ) -> Iterable[str]:
        """Suggest directive options that may be used, given a completion context."""

        if domain:
            name = f"{domain}:{directive}"
        else:
            name = directive

        for feature_name, feature in self._features.items():
            try:
                yield from feature.suggest_options(context, directive, domain)
            except Exception:
                self.logger.error(
                    "Unable to suggest options for directive '%s', error in feature: '%s'\n%s",
                    name,
                    feature_name,
                    traceback.format_exc(),
                )

    completion_triggers = [DIRECTIVE, DIRECTIVE_OPTION]
    definition_triggers = [DIRECTIVE]
    hover_triggers = [DIRECTIVE]
    implementation_triggers = [DIRECTIVE]

    def completion_resolve(self, item: CompletionItem) -> CompletionItem:
        # We need extra info to know who to call.
        if not item.data:
            return item

        data = typing.cast(Dict, item.data)
        ctype = data.get("completion_type", "")

        if ctype == "directive":
            return self.completion_resolve_directive(item)

        if ctype == "directive_option":
            return self.completion_resolve_option(item)

        return item

    def complete(self, context: CompletionContext) -> List[CompletionItem]:
        # Do not suggest completions within the middle of Python code.
        if context.location == "py":
            return []

        groups = context.match.groupdict()

        # Are we completing a directive's options?
        if "directive" not in groups:
            return self.complete_options(context)

        # Are we completing the directive's argument?
        directive_end = context.match.span()[0] + len(groups["directive"])
        complete_directive = groups["directive"].endswith("::")

        if complete_directive and directive_end < context.position.character:
            return self.complete_arguments(context)

        return self.complete_directives(context)

    def complete_arguments(self, context: CompletionContext) -> List[CompletionItem]:
        arguments = []
        name = context.match.group("name")
        domain = context.match.group("domain") or ""

        for feature in self._features.values():
            arguments += feature.complete_arguments(context, domain, name) or []

        return arguments

    def complete_directives(self, context: CompletionContext) -> List[CompletionItem]:
        self.logger.debug("Completing directives")

        items = []
        for name, directive in self.suggest_directives(context):
            item = render_directive_completion(context, name, directive)
            if item is None:
                continue

            items.append(item)

        return items

    def completion_resolve_directive(self, item: CompletionItem) -> CompletionItem:
        # We need the detail field set to the implementation's fully qualified name.
        if not item.detail:
            return item

        documentation = self.get_documentation(item.label, item.detail)
        if not documentation:
            return item

        description = documentation.get("description", "")
        is_markdown = documentation.get("is_markdown", False)
        kind = MarkupKind.Markdown if is_markdown else MarkupKind.PlainText

        item.documentation = MarkupContent(kind=kind, value=description)
        return item

    def complete_options(self, context: CompletionContext) -> List[CompletionItem]:
        surrounding_directive = self._get_surrounding_directive(context)
        if not surrounding_directive:
            return []

        name = surrounding_directive.group("name")
        domain = surrounding_directive.group("domain")
        impl = self.get_implementation(name, domain)
        if impl is None:
            return []

        items = []

        for option in self.suggest_options(context, name, domain):
            item = render_directive_option_completion(context, option, name, impl)
            if item is None:
                continue

            items.append(item)

        return items

    def completion_resolve_option(self, item: CompletionItem) -> CompletionItem:
        # We need the detail field set to the implementation's fully qualified name.
        if not item.detail or not item.data:
            return item

        directive, option = item.detail.split(":")
        name = typing.cast(Dict, item.data).get("for_directive", "")

        documentation = self.get_documentation(name, directive)
        if not documentation:
            return item

        description = documentation.get("options", {}).get(option, None)
        if not description:
            return item

        source = documentation.get("source", "")
        license = documentation.get("license", "")

        if source:
            description += f"\n\n[Source]({source})"

        if license:
            description += f"\n\n[License]({license})"

        kind = MarkupKind.PlainText
        if documentation.get("is_markdown", False):
            kind = MarkupKind.Markdown

        item.documentation = MarkupContent(kind=kind, value=description)
        return item

    def definition(self, context: DefinitionContext) -> List[Location]:
        directive = context.match.group("name")
        domain = context.match.group("domain")
        argument = context.match.group("argument")

        if not argument:
            return []

        start = context.match.group(0).index(argument)
        end = start + len(argument)

        if start <= context.position.character <= end:
            return self.find_argument_definition(context, directive, domain, argument)

        return []

    def find_argument_definition(
        self,
        context: DefinitionContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> List[Location]:
        definitions = []

        for feature_name, feature in self._features.items():
            try:
                definitions += (
                    feature.find_argument_definitions(
                        context, directive, domain, argument
                    )
                    or []
                )
            except Exception:
                self.logger.error(
                    "Unable to find definitions of '%s' for directive '%s', "
                    "error in feature: '%s'",
                    argument,
                    f"{domain}:{directive}" if domain else directive,
                    feature_name,
                    exc_info=True,
                )

        return definitions

    def resolve_argument_link(
        self, context: DocumentLinkContext, name: str, domain: str, argument: str
    ) -> Tuple[Optional[str], Optional[str]]:
        for feature_name, feature in self._features.items():
            try:
                target, tooltip = feature.resolve_argument_link(
                    context, name, domain, argument
                )

                if target:
                    return target, tooltip
            except Exception:
                self.logger.error(
                    "Unable to resolve argument link '%s' for directive '%s', "
                    "error in feature: '%s'",
                    argument,
                    f"{domain}:{name}" if domain else name,
                    feature_name,
                    exc_info=True,
                )

        return None, None

    def document_link(self, context: DocumentLinkContext) -> List[DocumentLink]:
        links = []

        for line, text in enumerate(context.doc.lines):
            for match in DIRECTIVE.finditer(text):
                argument = match.group("argument")
                if not argument:
                    continue

                domain = match.group("domain")
                name = match.group("name")

                target, tooltip = self.resolve_argument_link(
                    context, name, domain, argument
                )
                if not target:
                    continue

                idx = match.group(0).index(argument)
                start = match.start() + idx
                end = start + len(argument)

                links.append(
                    DocumentLink(
                        target=target,
                        tooltip=tooltip if context.tooltip_support else None,
                        range=Range(
                            start=Position(line=line, character=start),
                            end=Position(line=line, character=end),
                        ),
                    )
                )

        return links

    def hover(self, context: HoverContext) -> str:
        if context.location not in {"rst", "docstring"}:
            return ""

        name = context.match.group("name")
        domain = context.match.group("domain")

        # Determine if the hover is on the .. directive::  itself, or within the argument
        # Be sure to include enough chars for the length of '::'!
        idx = context.position.character - context.match.start()
        prefix = context.match.group(0)[:idx]

        if "::" not in prefix:
            return self.hover_directive(context, name, domain)

        # TODO: Add extension points for directive arguments and options.
        return ""

    def hover_directive(
        self, context: HoverContext, name: str, domain: Optional[str]
    ) -> str:
        label = f"{domain}:{name}" if domain else name
        self.logger.debug("Calculating hover for directive '%s'", label)

        directive = self.get_implementation(name, domain)
        if not directive:
            return ""

        try:
            dotted_name = f"{directive.__module__}.{directive.__name__}"
        except AttributeError:
            dotted_name = f"{directive.__module__}.{directive.__class__.__name__}"

        documentation = self.get_documentation(label, dotted_name)
        if not documentation:
            return ""

        return documentation.get("description", "")

    def implementation(self, context: ImplementationContext) -> List[Location]:
        region = context.match.group("directive")
        name = context.match.group("name")
        domain = context.match.group("domain")

        start = context.match.group(0).index(region)
        end = start + len(region)

        if start <= context.position.character <= end:
            return self.find_directive_implementation(context, name, domain)

        return []

    def find_directive_implementation(
        self, context: ImplementationContext, name: str, domain: Optional[str]
    ) -> List[Location]:
        impl = self.get_implementation(name, domain)
        if impl is None:
            return []

        self.logger.debug(
            "Getting implementation of '%s' (%s)",
            f"{domain}:{name}" if domain else name,
            impl,
        )
        location = get_object_location(impl, self.logger)
        if location is None:
            return []

        return [location]

    def _get_surrounding_directive(
        self, context: CompletionContext
    ) -> Optional["re.Match"]:
        """Used to determine which directive we should be offering completions for.

        When suggestions should be generated this returns an :class:`python:re.Match`
        object representing the directive the options are associated with. In the
        case where suggestions should not be generated this will return ``None``

        Parameters
        ----------
        context:
          The completion context
        """

        match = context.match
        groups = match.groupdict()
        indent = groups["indent"]

        self.logger.debug("Match groups: %s", groups)

        # Search backwards so that we can determine the context for our completion
        linum = context.position.line - 1
        line = context.doc.lines[linum]

        while linum >= 0 and line.startswith(indent):
            linum -= 1
            line = context.doc.lines[linum]

        # Only offer completions if we're within a directive's option block
        directive = DIRECTIVE.match(line)
        self.logger.debug("Context line:  %s", line)
        self.logger.debug("Context match: %s", directive)

        if not directive:
            return None

        # Now that we know we're in a directive's option block, is the completion
        # request coming from a valid position on the line?
        option = groups["option"]
        start = match.span()[0] + match.group(0).find(option)
        end = start + len(option) + 1

        if start <= context.position.character <= end:
            return directive

        return None

    def get_documentation(
        self, label: str, implementation: str
    ) -> Optional[Dict[str, Any]]:
        """Return the documentation for the given directive, if available.

        If documentation for the given ``label`` cannot be found, this function will also
        look for the label under the project's :confval:`sphinx:primary_domain` followed
        by the ``std`` domain.

        Parameters
        ----------
        label
           The name of the directive, as the user would type in an reStructuredText file.

        implementation
           The full dotted name of the directive's implementation.
        """

        key = f"{label}({implementation})"
        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        if not isinstance(self.rst, SphinxLanguageServer) or not self.rst.app:
            return None

        # Nothing found, try the primary domain
        domain = self.rst.app.config.primary_domain
        key = f"{domain}:{label}({implementation})"

        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        # Still nothing, try the standard domain
        key = f"std:{label}({implementation})"

        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        return None


def esbonio_setup(rst: RstLanguageServer):
    rst.add_feature(Directives(rst))



import enum
import importlib
import json
import logging
import textwrap
import traceback
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Type

from lsprotocol.types import COMPLETION_ITEM_RESOLVE
from lsprotocol.types import INITIALIZE
from lsprotocol.types import INITIALIZED
from lsprotocol.types import SHUTDOWN
from lsprotocol.types import TEXT_DOCUMENT_CODE_ACTION
from lsprotocol.types import TEXT_DOCUMENT_COMPLETION
from lsprotocol.types import TEXT_DOCUMENT_DEFINITION
from lsprotocol.types import TEXT_DOCUMENT_DID_CHANGE
from lsprotocol.types import TEXT_DOCUMENT_DID_OPEN
from lsprotocol.types import TEXT_DOCUMENT_DID_SAVE
from lsprotocol.types import TEXT_DOCUMENT_DOCUMENT_LINK
from lsprotocol.types import TEXT_DOCUMENT_DOCUMENT_SYMBOL
from lsprotocol.types import TEXT_DOCUMENT_HOVER
from lsprotocol.types import TEXT_DOCUMENT_IMPLEMENTATION
from lsprotocol.types import WORKSPACE_DID_DELETE_FILES
from lsprotocol.types import CodeActionParams
from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionList
from lsprotocol.types import CompletionOptions
from lsprotocol.types import CompletionParams
from lsprotocol.types import DefinitionParams
from lsprotocol.types import DeleteFilesParams
from lsprotocol.types import DidChangeTextDocumentParams
from lsprotocol.types import DidOpenTextDocumentParams
from lsprotocol.types import DidSaveTextDocumentParams
from lsprotocol.types import DocumentLinkParams
from lsprotocol.types import DocumentSymbolParams
from lsprotocol.types import FileOperationFilter
from lsprotocol.types import FileOperationPattern
from lsprotocol.types import FileOperationRegistrationOptions
from lsprotocol.types import Hover
from lsprotocol.types import HoverParams
from lsprotocol.types import ImplementationParams
from lsprotocol.types import InitializedParams
from lsprotocol.types import InitializeParams
from lsprotocol.types import MarkupContent
from lsprotocol.types import MarkupKind

from .rst import CompletionContext
from .rst import DefinitionContext
from .rst import DocumentLinkContext
from .rst import HoverContext
from .rst import ImplementationContext
from .rst import LanguageFeature
from .rst import RstLanguageServer
from .symbols import SymbolVisitor

__version__ = "0.16.4"

__all__ = [
    "CompletionContext",
    "DefinitionContext",
    "DocumentLinkContext",
    "HoverContext",
    "ImplementationContext",
    "LanguageFeature",
    "RstLanguageServer",
    "create_language_server",
]

logger = logging.getLogger(__name__)

# Commands
ESBONIO_SERVER_CONFIGURATION = "esbonio.server.configuration"
ESBONIO_SERVER_PREVIEW = "esbonio.server.preview"
ESBONIO_SERVER_BUILD = "esbonio.server.build"


def create_language_server(
    server_cls: Type[RstLanguageServer], modules: Iterable[str], *args, **kwargs
) -> RstLanguageServer:
    """Create a new language server instance.

    Parameters
    ----------
    server_cls:
       The class definition to create the server from.
    modules:
       The list of modules that should be loaded.
    args, kwargs:
       Any additional arguments that should be passed to the language server's
       constructor.
    """

    if "logger" not in kwargs:
        kwargs["logger"] = logger

    server = server_cls(*args, **kwargs)

    for module in modules:
        _load_module(server, module)

    return _configure_lsp_methods(server)


def _configure_lsp_methods(server: RstLanguageServer) -> RstLanguageServer:
    @server.feature(INITIALIZE)
    def on_initialize(ls: RstLanguageServer, params: InitializeParams):
        ls.initialize(params)

        for feature in ls._features.values():
            feature.initialize(params)

    @server.feature(INITIALIZED)
    def on_initialized(ls: RstLanguageServer, params: InitializedParams):
        ls.initialized(params)

        for feature in ls._features.values():
            feature.initialized(params)

    @server.feature(SHUTDOWN)
    def on_shutdown(ls: RstLanguageServer, *args):
        ls.on_shutdown(*args)

        for feature in ls._features.values():
            feature.on_shutdown(*args)

    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    def on_open(ls: RstLanguageServer, params: DidOpenTextDocumentParams):
        ...

    @server.feature(TEXT_DOCUMENT_DID_CHANGE)
    def on_change(ls: RstLanguageServer, params: DidChangeTextDocumentParams):
        pass

    @server.command(ESBONIO_SERVER_BUILD)
    def build(ls: RstLanguageServer, *args):
        params = {} if not args[0] else args[0][0]._asdict()
        force_all: bool = params.get("force_all", False)
        filenames: Optional[List[str]] = params.get("filenames", None)
        ls.build(force_all, filenames)

    @server.feature(TEXT_DOCUMENT_DID_SAVE)
    def on_save(ls: RstLanguageServer, params: DidSaveTextDocumentParams):
        ls.save(params)

        for feature in ls._features.values():
            feature.save(params)

    @server.feature(
        WORKSPACE_DID_DELETE_FILES,
        FileOperationRegistrationOptions(
            filters=[
                FileOperationFilter(
                    pattern=FileOperationPattern(glob="**/*.rst"),
                )
            ]
        ),
    )
    def on_delete_files(ls: RstLanguageServer, params: DeleteFilesParams):
        ls.delete_files(params)

        for feature in ls._features.values():
            feature.delete_files(params)

    @server.feature(TEXT_DOCUMENT_CODE_ACTION)
    def on_code_action(ls: RstLanguageServer, params: CodeActionParams):
        actions = []

        for feature in ls._features.values():
            actions += feature.code_action(params)

        return actions

    @server.feature(TEXT_DOCUMENT_HOVER)
    def on_hover(ls: RstLanguageServer, params: HoverParams):
        uri = params.text_document.uri
        doc = ls.workspace.get_document(uri)
        pos = params.position
        line = ls.line_at_position(doc, pos)
        location = ls.get_location_type(doc, pos)

        hover_values = []
        for feature in ls._features.values():
            for pattern in feature.hover_triggers:
                for match in pattern.finditer(line):
                    if not match:
                        continue

                    # only trigger hover if the position of the request is within
                    # the match
                    start, stop = match.span()
                    if start <= pos.character <= stop:
                        context = HoverContext(
                            doc=doc,
                            location=location,
                            match=match,
                            position=pos,
                            capabilities=ls.client_capabilities,
                        )
                        ls.logger.debug("Hover context: %s", context)

                        hover_value = feature.hover(context)
                        hover_values.append(hover_value)

        hover_content_values = "\n".join(hover_values)

        return Hover(
            contents=MarkupContent(
                kind=MarkupKind.Markdown,
                value=hover_content_values,
            )
        )

    # <engine-example>
    @server.feature(
        TEXT_DOCUMENT_COMPLETION,
        CompletionOptions(
            trigger_characters=[">", ".", ":", "`", "<", "/"], resolve_provider=True
        ),
    )
    def on_completion(ls: RstLanguageServer, params: CompletionParams):
        uri = params.text_document.uri
        pos = params.position

        doc = ls.workspace.get_document(uri)
        line = ls.line_at_position(doc, pos)
        location = ls.get_location_type(doc, pos)

        items = []

        for name, feature in ls._features.items():
            for pattern in feature.completion_triggers:
                for match in pattern.finditer(line):
                    if not match:
                        continue

                    # Only trigger completions if the position of the request is within
                    # the match.
                    start, stop = match.span()
                    if start <= pos.character <= stop:
                        context = CompletionContext(
                            doc=doc,
                            location=location,
                            match=match,
                            position=pos,
                            config=ls.user_config.server.completion,
                            capabilities=ls.client_capabilities,
                        )
                        ls.logger.debug("Completion context: %s", context)

                        for item in feature.complete(context):
                            item.data = {"source_feature": name, **(item.data or {})}  # type: ignore
                            items.append(item)

        return CompletionList(is_incomplete=False, items=items)

    # </engine-example>

    @server.feature(COMPLETION_ITEM_RESOLVE)
    def on_completion_resolve(
        ls: RstLanguageServer, item: CompletionItem
    ) -> CompletionItem:
        source = (item.data or {}).get("source_feature", "")  # type: ignore
        feature = ls.get_feature(source)

        if not feature:
            ls.logger.error(
                "Unable to resolve completion item, unknown source: '%s'", source
            )
            return item

        return feature.completion_resolve(item)

    @server.feature(TEXT_DOCUMENT_DEFINITION)
    def on_definition(ls: RstLanguageServer, params: DefinitionParams):
        uri = params.text_document.uri
        pos = params.position

        doc = ls.workspace.get_document(uri)
        line = ls.line_at_position(doc, pos)
        location = ls.get_location_type(doc, pos)

        definitions = []

        for feature in ls._features.values():
            for pattern in feature.definition_triggers:
                for match in pattern.finditer(line):
                    if not match:
                        continue

                    start, stop = match.span()
                    if start <= pos.character and pos.character <= stop:
                        context = DefinitionContext(
                            doc=doc, location=location, match=match, position=pos
                        )
                        definitions += feature.definition(context)

        return definitions

    @server.feature(TEXT_DOCUMENT_IMPLEMENTATION)
    def on_implementation(ls: RstLanguageServer, params: ImplementationParams):
        uri = params.text_document.uri
        pos = params.position

        doc = ls.workspace.get_document(uri)
        line = ls.line_at_position(doc, pos)
        location = ls.get_location_type(doc, pos)

        implementations = []

        for feature in ls._features.values():
            for pattern in feature.implementation_triggers:
                for match in pattern.finditer(line):
                    if not match:
                        continue

                    start, stop = match.span()
                    if start <= pos.character and pos.character <= stop:
                        context = ImplementationContext(
                            doc=doc, location=location, match=match, position=pos
                        )
                        ls.logger.debug("Implementation context: %s", context)
                        implementations += feature.implementation(context)

        return implementations

    @server.feature(TEXT_DOCUMENT_DOCUMENT_LINK)
    def on_document_link(ls: RstLanguageServer, params: DocumentLinkParams):
        uri = params.text_document.uri
        doc = ls.workspace.get_document(uri)
        context = DocumentLinkContext(doc=doc, capabilities=ls.client_capabilities)

        links = []
        for feature in ls._features.values():
            links += feature.document_link(context) or []

        return links

    @server.feature(TEXT_DOCUMENT_DOCUMENT_SYMBOL)
    def on_document_symbol(ls: RstLanguageServer, params: DocumentSymbolParams):
        doctree = ls.get_initial_doctree(uri=params.text_document.uri)
        if doctree is None:
            return None

        visitor = SymbolVisitor(ls, doctree)
        doctree.walkabout(visitor)

        return visitor.symbols

    @server.command(ESBONIO_SERVER_CONFIGURATION)
    def get_configuration(ls: RstLanguageServer, *args):
        """Get the server's configuration.

        Not to be confused with the ``workspace/configuration`` request where the server
        can request the client's configuration. This is so clients can ask for sphinx's
        output path for example.

        As far as I know, there isn't anything built into the spec to cater for this?
        """
        config = ls.configuration
        ls.logger.debug("%s: %s", ESBONIO_SERVER_CONFIGURATION, dump(config))

        return config

    @server.command(ESBONIO_SERVER_PREVIEW)
    def preview(ls: RstLanguageServer, *args) -> Dict[str, Any]:
        """Start/Generate a preview of the project"""
        params = {} if not args[0] else args[0][0]
        ls.logger.debug("%s: %s", ESBONIO_SERVER_PREVIEW, params)

        return ls.preview(params) or {}

    return server


def _load_module(server: RstLanguageServer, modname: str):
    """Load an extension module by calling its ``esbonio_setup`` function, if it exists."""

    try:
        module = importlib.import_module(modname)
    except ImportError:
        logger.error(
            "Unable to import module '%s'\n%s", modname, traceback.format_exc()
        )
        return None

    setup = getattr(module, "esbonio_setup", None)
    if setup is None:
        logger.error("Skipping module '%s', missing 'esbonio_setup' function", modname)
        return None

    server.load_extension(modname, setup)


def dump(obj) -> str:
    """Debug helper function that converts an object to JSON."""

    def default(o):
        if isinstance(o, enum.Enum):
            return o.value

        fields = {}
        for k, v in o.__dict__.items():
            if v is None:
                continue

            # Truncate long strings - but not uris!
            if isinstance(v, str) and not k.lower().endswith("uri"):
                v = textwrap.shorten(v, width=25)

            fields[k] = v

        return fields

    return json.dumps(obj, default=default)



from typing import Optional

from docutils import nodes
from docutils.nodes import NodeVisitor
from lsprotocol.types import DocumentSymbol
from lsprotocol.types import Position
from lsprotocol.types import Range
from lsprotocol.types import SymbolKind


class SymbolVisitor(NodeVisitor):
    """A visitor used to build the hierarchy we return from a
    ``textDocument/documentSymbol`` request.

    """

    def __init__(self, rst, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.logger = rst.logger
        self.symbols = []
        self.symbol_stack = []

    @property
    def current_symbol(self) -> Optional[DocumentSymbol]:
        if len(self.symbol_stack) == 0:
            return None

        return self.symbol_stack[-1]

    def push_symbol(self):
        symbol = DocumentSymbol(
            name="",
            kind=SymbolKind.String,
            range=Range(
                start=Position(line=1, character=0),
                end=Position(line=1, character=10),
            ),
            selection_range=Range(
                start=Position(line=1, character=0),
                end=Position(line=1, character=10),
            ),
            children=[],
        )
        current_symbol = self.current_symbol

        if not current_symbol:
            self.symbols.append(symbol)
        else:
            if current_symbol.children is None:
                current_symbol.children = [symbol]
            else:
                current_symbol.children.append(symbol)

        self.symbol_stack.append(symbol)
        return symbol

    def pop_symbol(self):
        self.symbol_stack.pop()

    def visit_section(self, node: nodes.Node) -> None:
        self.push_symbol()

    def depart_section(self, node: nodes.Node) -> None:
        self.pop_symbol()

    def visit_title(self, node: nodes.Node) -> None:
        symbol = self.current_symbol
        has_parent = True

        if symbol is None:
            has_parent = False
            symbol = self.push_symbol()

        name = node.astext()
        if len(name) == 0:
            name = "title_node"

        line = (node.line or 1) - 1

        symbol.name = name
        symbol.range.start.line = line
        symbol.range.end.line = line
        symbol.range.end.character = len(name) - 1
        symbol.selection_range.start.line = line
        symbol.selection_range.end.line = line
        symbol.selection_range.end.character = len(name) - 1

        if not has_parent:
            self.pop_symbol()

    def depart_title(self, node: nodes.Node) -> None:
        pass

    def visit_a_directive(self, node: nodes.Element):
        symbol = self.push_symbol()

        name = node["text"]  # type: ignore
        if len(name) == 0:
            name = "a_directive"

        line = (node.line or 1) - 1

        symbol.name = name
        symbol.kind = SymbolKind.Class
        symbol.range.start.line = line
        symbol.range.end.line = line
        symbol.range.end.character = len(name) - 1
        symbol.selection_range.start.line = line
        symbol.selection_range.end.line = line
        symbol.selection_range.end.character = len(name) - 1

    def depart_a_directive(self, node: nodes.Node):
        self.pop_symbol()

    # TODO: Enable symbols for roles
    #       However the reported line numbers can be inaccurate...
    def visit_a_role(self, node: nodes.Node) -> None:
        ...

    def depart_a_role(self, node: nodes.Node) -> None:
        ...

    # TODO: Enable symbols for definition list items
    #       However the reported line numbers appear to be inaccurate...

    def visit_Text(self, node: nodes.Node) -> None:
        pass

    def depart_Text(self, node: nodes.Node) -> None:
        pass

    def unknown_visit(self, node: nodes.Node) -> None:
        pass

    def unknown_departure(self, node: nodes.Node) -> None:
        pass



from ._patterns import DIRECTIVE
from ._patterns import ROLE
from ._uri import Uri
from .feature import LanguageFeature
from .server import EsbonioLanguageServer

__all__ = [
    "DIRECTIVE",
    "ROLE",
    "EsbonioLanguageServer",
    "LanguageFeature",
    "Uri",
]



import dataclasses
import json
import logging
import sys
from typing import Any
from typing import Union

logger = logging.getLogger("esbonio.sphinx_agent")


def format_message(data: Any) -> str:
    if dataclasses.is_dataclass(data):
        data = dataclasses.asdict(data)

    content = json.dumps(data)
    content_length = len(content)

    return f"Content-Length: {content_length}\r\n\r\n{content}"


def send_error(id: Union[str, int], code: int, message: str, data=None):
    send_message(
        dict(
            id=id,
            jsonrpc="2.0",
            error=dict(code=code, message=message, data=data),
        )
    )


def send_message(data: Any):
    content = format_message(data).encode("utf8")
    sys.stdout.buffer.write(content)
    sys.stdout.flush()



"""Subprocess implementation of the ``SphinxClient`` protocol."""
from __future__ import annotations

import asyncio
import json
import logging
import os
import sys
import typing
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pygls import IS_WIN
from pygls.client import JsonRPCClient
from pygls.protocol import JsonRPCProtocol

import esbonio.sphinx_agent.types as types
from esbonio.server import Uri

from .config import SphinxConfig

if typing.TYPE_CHECKING:
    from .client import SphinxClient
    from .manager import SphinxManager


class SphinxAgentProtocol(JsonRPCProtocol):
    """Describes the protocol spoken between the client below and the sphinx agent."""

    def get_message_type(self, method: str) -> Any | None:
        return types.METHOD_TO_MESSAGE_TYPE.get(method, None)

    def get_result_type(self, method: str) -> Any | None:
        return types.METHOD_TO_RESPONSE_TYPE.get(method, None)


class SubprocessSphinxClient(JsonRPCClient):
    """JSON-RPC client used to drive a Sphinx application instance hosted in
    a separate subprocess.

    See :mod:`esbonio.sphinx_agent` for the implementation of the server component.
    """

    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
        protocol_cls=SphinxAgentProtocol,
        *args,
        **kwargs,
    ):
        super().__init__(protocol_cls=protocol_cls, *args, **kwargs)  # type: ignore[misc]
        self.logger = logger or logging.getLogger(__name__)

        self.sphinx_info: Optional[types.SphinxInfo] = None
        self._building = False
        self._build_file_map: Dict[Uri, str] = {}
        self._diagnostics: Dict[Uri, List[types.Diagnostic]] = {}

    @property
    def id(self) -> Optional[str]:
        """The id of the Sphinx instance."""
        if self.sphinx_info is None:
            return None

        return self.sphinx_info.id

    @property
    def building(self) -> bool:
        return self._building

    @property
    def builder(self) -> Optional[str]:
        """The sphinx application's builder name"""
        if self.sphinx_info is None:
            return None

        return self.sphinx_info.builder_name

    @property
    def src_uri(self) -> Optional[Uri]:
        """The src uri of the Sphinx application."""
        if self.sphinx_info is None:
            return None

        return Uri.for_file(self.sphinx_info.src_dir)

    @property
    def conf_uri(self) -> Optional[Uri]:
        """The conf uri of the Sphinx application."""
        if self.sphinx_info is None:
            return None

        return Uri.for_file(self.sphinx_info.conf_dir)

    @property
    def diagnostics(self) -> Dict[Uri, List[types.Diagnostic]]:
        """Any diagnostics associated with the project.

        These are automatically updated with each build.
        """
        return self._diagnostics

    @property
    def build_file_map(self) -> Dict[Uri, str]:
        """Mapping of source files to their corresponing output path."""
        return self._build_file_map

    @property
    def build_uri(self) -> Optional[Uri]:
        """The build uri of the Sphinx application."""
        if self.sphinx_info is None:
            return None

        return Uri.for_file(self.sphinx_info.build_dir)

    async def server_exit(self, server: asyncio.subprocess.Process):
        """Called when the sphinx agent process exits."""

        #   0: all good
        # -15: terminated
        if server.returncode not in {0, -15}:
            self.logger.error(
                f"sphinx-agent process exited with code: {server.returncode}"
            )

            if server.stderr is not None:
                stderr = await server.stderr.read()
                self.logger.error("Stderr:\n%s", stderr.decode("utf8"))

        # TODO: Should the upstream base client be doing this?
        # Cancel any pending futures.
        for id_, fut in self.protocol._request_futures.items():
            message = "Cancelled" if fut.cancel() else "Unable to cancel"
            self.logger.debug(
                "%s future '%s' for pending request '%s'", message, fut, id_
            )

    async def start(self, config: SphinxConfig):
        """Start the client."""
        command = []
        if config.enable_dev_tools:
            command.extend([sys.executable, "-m", "lsp_devtools", "agent", "--"])

        command.extend([*config.python_command, "-m", "sphinx_agent"])
        env = get_sphinx_env(config)

        self.logger.debug("Sphinx agent env: %s", json.dumps(env, indent=2))
        self.logger.debug("Starting sphinx agent: %s", " ".join(command))

        await self.start_io(*command, env=env, cwd=config.cwd)

    async def stop(self):
        """Stop the client."""
        self.protocol.notify("exit", None)

        # Give the agent a little time to close.
        await asyncio.sleep(0.5)
        await super().stop()

    async def create_application(self, config: SphinxConfig) -> types.SphinxInfo:
        """Create a sphinx application object."""

        params = types.CreateApplicationParams(
            command=config.build_command,
            enable_sync_scrolling=config.enable_sync_scrolling,
        )

        sphinx_info = await self.protocol.send_request_async("sphinx/createApp", params)
        self.sphinx_info = sphinx_info
        return sphinx_info

    async def build(
        self,
        *,
        filenames: Optional[List[str]] = None,
        force_all: bool = False,
        content_overrides: Optional[Dict[str, str]] = None,
    ) -> types.BuildResult:
        """Trigger a Sphinx build."""

        params = types.BuildParams(
            filenames=filenames or [],
            force_all=force_all,
            content_overrides=content_overrides or {},
        )

        self._building = True

        result = await self.protocol.send_request_async("sphinx/build", params)
        self._building = False

        self._diagnostics = {
            Uri.for_file(fpath): items for fpath, items in result.diagnostics.items()
        }

        self._build_file_map = {
            Uri.for_file(src): out for src, out in result.build_file_map.items()
        }

        return result


def make_subprocess_sphinx_client(manager: SphinxManager) -> SphinxClient:
    """Factory function for creating a ``SubprocessSphinxClient`` instance.

    Parameters
    ----------
    manager
       The manager instance creating the client

    Returns
    -------
    SphinxClient
       The configured client
    """
    client = SubprocessSphinxClient(logger=manager.logger)

    @client.feature("window/logMessage")
    def _on_msg(ls: SubprocessSphinxClient, params):
        manager.server.show_message_log(params.message)

    return client


def make_test_sphinx_client() -> SubprocessSphinxClient:
    """Factory function for creating a ``SubprocessSphinxClient`` instance
    to use for testing."""
    logger = logging.getLogger("sphinx_client")
    logger.setLevel(logging.INFO)

    client = SubprocessSphinxClient()

    @client.feature("window/logMessage")
    def _(params):
        logger.info("%s", params.message)

    return client


def get_sphinx_env(config: SphinxConfig) -> Dict[str, str]:
    """Return the set of environment variables to use with the Sphinx process."""
    env = {"PYTHONPATH": ":".join([str(p) for p in config.python_path])}

    passthrough = set(config.env_passthrough)
    if IS_WIN and "SYSTEMROOT" not in passthrough:
        passthrough.add("SYSTEMROOT")

    for envname in passthrough:
        value = os.environ.get(envname, None)
        if value is not None:
            if envname == "PYTHONPATH":
                env["PYTHONPATH"] = f"{env['PYTHONPATH']}:{value}"
            else:
                env[envname] = value

    return env



"""Default entry point, identical to calling ``python -m esbonio.lsp.sphinx``"""
import sys

from esbonio import cli
from esbonio.lsp.sphinx import DEFAULT_MODULES
from esbonio.lsp.sphinx import SphinxLanguageServer

_cli = cli.setup_cli("esbonio", "Esbonio's Sphinx language server.")
_cli.set_defaults(modules=DEFAULT_MODULES)
_cli.set_defaults(server_cls=SphinxLanguageServer)


def main():
    cli.main(_cli)


if __name__ == "__main__":
    sys.exit(main())



import asyncio
import logging
from http.server import HTTPServer
from http.server import SimpleHTTPRequestHandler
from typing import Any
from typing import Dict
from typing import Optional
from urllib.parse import urlencode

import attrs

from esbonio.server import EsbonioLanguageServer
from esbonio.server import Uri
from esbonio.server.feature import LanguageFeature
from esbonio.server.features.sphinx_manager import SphinxClient
from esbonio.server.features.sphinx_manager import SphinxManager

from .webview import WebviewServer
from .webview import make_ws_server


class RequestHandler(SimpleHTTPRequestHandler):
    def __init__(self, *args, logger: logging.Logger, directory: str, **kwargs) -> None:
        self.logger = logger
        super().__init__(*args, directory=directory, **kwargs)

    def translate_path(self, path: str) -> str:
        result = super().translate_path(path)
        # self.logger.debug("Translate: '%s' -> '%s'", path, result)
        return result

    def log_message(self, format: str, *args: Any) -> None:
        self.logger.debug(format, *args)


class RequestHandlerFactory:
    """Class for dynamically producing request handlers.

    ``HTTPServer`` works by taking a "request handler" class and creating an instance of
    it for every request it receives. By making this class callable, we can dynamically
    produce a request handler based on the current situation.
    """

    def __init__(self, logger: logging.Logger, build_uri: Optional[Uri] = None):
        self.logger = logger
        self.build_uri = build_uri

    def __call__(self, *args, **kwargs):
        if self.build_uri is None:
            raise ValueError("No build directory set")

        if (build_dir := self.build_uri.fs_path) is None:
            raise ValueError(
                "Unable to determine build dir from uri: '%s'", self.build_uri
            )

        return RequestHandler(*args, logger=self.logger, directory=build_dir, **kwargs)


@attrs.define
class PreviewConfig:
    """Configuration settings for previews."""

    bind: str = attrs.field(default="localhost")
    """The network interface to bind to, defaults to ``localhost``"""

    http_port: int = attrs.field(default=0)
    """The port to host the HTTP server on. If ``0`` a random port number will be
    chosen"""

    ws_port: int = attrs.field(default=0)
    """The port to host the WebSocket server on. If ``0`` a random port number will be
    chosen"""

    show_line_markers: bool = attrs.field(default=False)
    """If set, render the source line markers in the preview"""


class PreviewManager(LanguageFeature):
    """Language feature for managing previews."""

    def __init__(self, server: EsbonioLanguageServer, sphinx: SphinxManager):
        super().__init__(server)
        self.sphinx = sphinx
        self.sphinx.add_listener("build", self.on_build)

        logger = server.logger.getChild("PreviewServer")
        self._request_handler_factory = RequestHandlerFactory(logger)
        self._http_server: Optional[HTTPServer] = None
        self._http_future: Optional[asyncio.Future] = None

        self._ws_server: Optional[WebviewServer] = None
        self._ws_task: Optional[asyncio.Task] = None

    @property
    def preview_active(self) -> bool:
        """Return true if the preview is active.

        i.e. there is a HTTP server hosting the build result."""
        return self._http_server is not None

    @property
    def preview_controllable(self) -> bool:
        """Return true if the preview is controllable.

        i.e. there is a web socket server available to control the webview.
        """
        return self._ws_server is not None

    async def get_preview_config(self) -> PreviewConfig:
        """Return the user's preview server configuration."""
        config = await self.server.get_user_config("esbonio.preview", PreviewConfig)
        if config is None:
            self.logger.info(
                "Unable to obtain preview configuration, proceeding with defaults"
            )
            config = PreviewConfig()

        return config

    async def get_http_server(self, config: PreviewConfig) -> HTTPServer:
        """Return the http server instance hosting the previews.

        This will also handle the creation of the server the first time it is called.
        """
        # TODO: Recreate the server if the configuration changes?
        if self._http_server is not None:
            return self._http_server

        self._http_server = HTTPServer(
            (config.bind, config.http_port), self._request_handler_factory
        )

        loop = asyncio.get_running_loop()
        self._http_future = loop.run_in_executor(
            self.server.thread_pool_executor,
            self._http_server.serve_forever,
        )

        return self._http_server

    async def get_webview_server(self, config: PreviewConfig) -> WebviewServer:
        """Return the websocket server used to communicate with the webview."""

        # TODO: Recreate the server if the configuration changes?
        if self._ws_server is not None:
            return self._ws_server

        logger = self.server.logger.getChild("WebviewServer")
        self._ws_server = make_ws_server(self.server, logger)
        self._ws_task = asyncio.create_task(
            self._ws_server.start_ws(config.bind, config.ws_port)
        )

        # HACK: we need to yield control to the event loop to give the ws_server time to
        #       spin up and allocate a port number.
        await asyncio.sleep(1)

        return self._ws_server

    async def on_build(self, client: SphinxClient, result):
        """Called whenever a sphinx build completes."""

        if self._ws_server is None:
            return

        # Only refresh the view if the project we are previewing was built.
        if client.build_uri != self._request_handler_factory.build_uri:
            return

        self.logger.debug("Refreshing preview")
        self._ws_server.reload()

    async def scroll_view(self, line: int):
        """Scroll the webview to the given line number."""

        if self._ws_server is None:
            return

        self._ws_server.scroll(line)

    async def preview_file(self, params):
        # Always check the fully resolved uri.
        src_uri = Uri.parse(params["uri"]).resolve()
        self.logger.debug("Previewing file: '%s'", src_uri)

        client = await self.sphinx.get_client(src_uri)
        if client is None:
            return None

        if client.builder not in {"html", "dirhtml"}:
            self.logger.error(
                "Previews for the '%s' builder are not currently supported",
                client.builder,
            )
            return None

        if (build_path := client.build_file_map.get(src_uri, None)) is None:
            # Has the project been built yet?
            if len(client.build_file_map) == 0:
                # If not, trigger a build and try again
                await client.build()
                return await self.preview_file(params)

            return None

        config = await self.get_preview_config()
        server = await self.get_http_server(config)
        webview = await self.get_webview_server(config)

        self._request_handler_factory.build_uri = client.build_uri
        query_params: Dict[str, Any] = dict(ws=webview.port)

        if config.show_line_markers:
            query_params["show-markers"] = True

        uri = Uri.create(
            scheme="http",
            authority=f"localhost:{server.server_port}",
            path=build_path,
            query=urlencode(query_params),
        )

        self.logger.info("Preview available at: %s", uri.as_string(encode=False))
        return {"uri": uri.as_string(encode=False)}


def esbonio_setup(server: EsbonioLanguageServer, sphinx: SphinxManager):
    manager = PreviewManager(server, sphinx)
    server.add_feature(manager)

    @server.feature("view/scroll")
    async def on_scroll(ls: EsbonioLanguageServer, params):
        await manager.scroll_view(params.line)

    @server.command("esbonio.server.previewFile")
    async def preview_file(ls: EsbonioLanguageServer, *args):
        return await manager.preview_file(args[0][0])



from __future__ import annotations

import importlib
import inspect
import typing
from typing import Iterable
from typing import Type

from lsprotocol import types

if typing.TYPE_CHECKING:
    from .server import EsbonioLanguageServer


def create_language_server(
    server_cls: Type[EsbonioLanguageServer], modules: Iterable[str], *args, **kwargs
) -> EsbonioLanguageServer:
    """Create a new language server instance.

    Parameters
    ----------
    server_cls:
       The class definition to create the server from.
    modules:
       The list of modules that should be loaded.
    args, kwargs:
       Any additional arguments that should be passed to the language server's
       constructor.
    """
    server = server_cls(*args, **kwargs)

    for module in modules:
        _load_module(server, module)

    return _configure_lsp_methods(server)


def _configure_lsp_methods(server: EsbonioLanguageServer) -> EsbonioLanguageServer:
    """Configure method handlers for the portions of the LSP spec we support."""

    @server.feature(types.INITIALIZE)
    async def on_initialize(ls: EsbonioLanguageServer, params: types.InitializeParams):
        ls.initialize(params)

    @server.feature(types.TEXT_DOCUMENT_DID_CHANGE)
    async def on_document_change(
        ls: EsbonioLanguageServer, params: types.DidChangeTextDocumentParams
    ):
        await call_features(ls, "document_change", params)

    @server.feature(types.TEXT_DOCUMENT_DID_CLOSE)
    async def on_document_close(
        ls: EsbonioLanguageServer, params: types.DidCloseTextDocumentParams
    ):
        await call_features(ls, "document_close", params)

    @server.feature(types.TEXT_DOCUMENT_DID_OPEN)
    async def on_document_open(
        ls: EsbonioLanguageServer, params: types.DidOpenTextDocumentParams
    ):
        await call_features(ls, "document_open", params)

    @server.feature(types.TEXT_DOCUMENT_DID_SAVE)
    async def on_document_save(
        ls: EsbonioLanguageServer, params: types.DidSaveTextDocumentParams
    ):
        # Record the version number of the document
        doc = ls.workspace.get_document(params.text_document.uri)
        doc.saved_version = doc.version or 0

        await call_features(ls, "document_save", params)

    @server.feature(types.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
    async def on_document_symbol(
        ls: EsbonioLanguageServer, params: types.DocumentSymbolParams
    ):
        return await call_features_return_first(ls, "document_symbol", params)

    return server


async def call_features(ls: EsbonioLanguageServer, method: str, *args, **kwargs):
    """Call all features."""

    for cls, feature in ls:
        try:
            impl = getattr(feature, method)

            result = impl(*args, **kwargs)
            if inspect.isawaitable(result):
                await result

        except Exception:
            name = f"{cls.__name__}"
            ls.logger.error("Error in '%s.%s' handler", name, method, exc_info=True)


async def call_features_return_first(
    ls: EsbonioLanguageServer, method: str, *args, **kwargs
):
    """Call all features, returning the first non ``None`` result we find."""

    for cls, feature in ls:
        try:
            impl = getattr(feature, method)

            result = impl(*args, **kwargs)
            if inspect.isawaitable(result):
                await result

            if result is not None:
                return result

        except Exception:
            name = f"{cls.__name__}"
            ls.logger.error("Error in '%s.%s' handler", name, method, exc_info=True)


def _load_module(server: EsbonioLanguageServer, modname: str):
    """Load an extension module by calling its ``esbonio_setup`` function, if it exists."""

    try:
        module = importlib.import_module(modname)
    except ImportError:
        server.logger.error("Unable to import module '%s'", modname, exc_info=True)
        return None

    setup = getattr(module, "esbonio_setup", None)
    if setup is None:
        server.logger.debug(
            "Skipping module '%s', missing 'esbonio_setup' function", modname
        )
        return None

    server.load_extension(modname, setup)



import logging
import typing
from typing import IO
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type

import pygls.uris as uri
from docutils import nodes
from docutils.core import Publisher
from docutils.io import NullOutput
from docutils.io import StringInput
from docutils.parsers.rst import Directive
from docutils.parsers.rst import Parser
from docutils.parsers.rst import directives
from docutils.parsers.rst import roles
from docutils.readers.standalone import Reader
from docutils.utils import Reporter
from docutils.writers import Writer
from pygls.workspace import Document
from sphinx.environment import default_settings

from esbonio.lsp.util.patterns import DIRECTIVE
from esbonio.lsp.util.patterns import ROLE


class a_directive(nodes.Element, nodes.Inline):
    """Represents a directive."""


class a_role(nodes.Element):
    """Represents a role."""

    def astext(self):
        return self["text"]


def dummy_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    node = a_role()
    node.line = lineno

    match = ROLE.match(rawtext)
    if match is None:
        node.attributes["text"] = rawtext
    else:
        node.attributes.update(match.groupdict())
        node.attributes["text"] = match.group(0)

    return [node], []


class DummyDirective(Directive):
    has_content = True

    def run(self):
        node = a_directive()
        node.line = self.lineno
        parent = self.state.parent
        lines = self.block_text

        # substitution definitions require special handling
        if isinstance(parent, nodes.substitution_definition):
            lines = parent.rawsource

        text = lines.split("\n")[0]
        match = DIRECTIVE.match(text)
        if match:
            node.attributes.update(match.groupdict())
            node.attributes["text"] = match.group(0)
        else:
            self.state.reporter.warning(f"Unable to parse directive: '{text}'")
            node.attributes["text"] = text

        if self.content:
            # This is essentially what `nested_parse_with_titles` does in Sphinx.
            # But by passing the content_offset to state.nested_parse we ensure any line
            # numbers remain relative to the start of the current file.
            current_titles = self.state.memo.title_styles
            current_sections = self.state.memo.section_level
            self.state.memo.title_styles = []
            self.state.memo.section_level = 0
            try:
                self.state.nested_parse(
                    self.content, self.content_offset, node, match_titles=1
                )
            finally:
                self.state.memo.title_styles = current_titles
                self.state.memo.section_level = current_sections

        return [node]


class disable_roles_and_directives:
    """Disables all roles and directives from being expanded.

    The ``CustomReSTDispactcher`` from Sphinx is *very* cool.
    It provides a way to override the mechanism used to lookup roles and directives
    during parsing!

    It's perfect for temporarily replacing all role and directive implementations with
    dummy ones. Parsing an rst document with these dummy implementations effectively
    gives us something that could be called an abstract syntax tree.

    Unfortunately, it's only available in a relatively recent version of Sphinx (4.4)
    so we have to implement the same idea ourselves for now.
    """

    def __init__(self) -> None:
        self.directive_backup: Optional[Callable] = None
        self.role_backup: Optional[Callable] = None

    def __enter__(self) -> None:
        self.directive_backup = directives.directive
        self.role_backup = roles.role

        directives.directive = self.directive
        roles.role = self.role

    def __exit__(self, exc_type: Type[Exception], exc_value: Exception, traceback: Any):
        directives.directive = self.directive_backup  # type: ignore
        roles.role = self.role_backup  # type: ignore

        self.directive_backup = None
        self.role_backup = None

    def directive(self, directive_name, language_module, document):
        return DummyDirective, []

    def role(self, role_name, language_module, lineno, reporter):
        return dummy_role, []


class DummyWriter(Writer):
    """A writer that doesn't do anything."""

    def translate(self):
        pass


class LogStream:
    def __init__(self, logger: logging.Logger):
        self.logger = logger

    def write(self, text: str):
        self.logger.debug(text)


class LogReporter(Reporter):
    """A docutils reporter that writes to the given logger."""

    def __init__(
        self,
        logger: logging.Logger,
        source: str,
        report_level: int,
        halt_level: int,
        debug: bool,
        error_handler: str,
    ) -> None:
        stream = typing.cast(IO, LogStream(logger))
        super().__init__(
            source, report_level, halt_level, stream, debug, error_handler=error_handler
        )  # type: ignore


class InitialDoctreeReader(Reader):
    """A reader that replaces the default reporter with one compatible with esbonio's
    logging setup."""

    def __init__(self, logger: logging.Logger, *args, **kwargs):
        self.logger = logger
        super().__init__(*args, **kwargs)

    def new_document(self) -> nodes.document:
        document = super().new_document()

        reporter = document.reporter
        document.reporter = LogReporter(
            self.logger,
            reporter.source,
            reporter.report_level,
            reporter.halt_level,
            reporter.debug_flag,
            reporter.error_handler,
        )

        return document


def read_initial_doctree(
    document: Document, logger: logging.Logger
) -> Optional[nodes.document]:
    """Parse the given reStructuredText document into its "initial" doctree.

    An "initial" doctree can be thought of as the abstract syntax tree of a
    reStructuredText document. This method disables all role and directives
    from being executed, instead they are replaced with nodes that simply
    represent that they exist.

    Parameters
    ----------
    document
       The document containing the reStructuredText source.

    logger
       Logger to log debug info to.
    """

    parser = Parser()
    with disable_roles_and_directives():
        publisher = Publisher(
            reader=InitialDoctreeReader(logger),
            parser=parser,
            writer=DummyWriter(),
            source_class=StringInput,
            destination=NullOutput(),
        )
        publisher.process_programmatic_settings(None, default_settings, None)
        publisher.set_source(
            source=document.source, source_path=uri.to_fs_path(document.uri)
        )
        publisher.publish()

        return publisher.document



import re
from typing import Any
from typing import Optional

from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionItemKind
from lsprotocol.types import Position
from lsprotocol.types import Range
from lsprotocol.types import TextEdit

from esbonio.lsp import CompletionContext

__all__ = ["render_role_completion"]


WORD = re.compile("[a-zA-Z]+")


def render_role_completion(
    context: CompletionContext,
    name: str,
    role: Any,
) -> Optional[CompletionItem]:
    """Render the given role as a ``CompletionItem`` according to the current
    context.

    Parameters
    ----------
    context
       The context in which the completion should be rendered.

    name
       The name of the role, as it appears in an rst file.

    role
       The implementation of the role.

    Returns
    -------
    Optional[CompletionItem]
       The final completion item or ``None``.
       If ``None``, then the given completion should be skipped.
    """

    if context.config.preferred_insert_behavior == "insert":
        return _render_role_with_insert_text(context, name, role)

    return _render_role_with_text_edit(context, name, role)


def _render_role_with_insert_text(context, name, role):
    """Render a role's ``CompletionItem`` using the ``insertText`` field.

    This implements the ``insert`` insert behavior for roles.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    name
       The name of the role, as it appears in an rst file.

    role
       The implementation of the role.
    """

    insert_text = f":{name}:"
    user_text = context.match.group(0).strip()

    if not insert_text.startswith(user_text):
        return None

    # If the existing text ends with a delimiter, then we should simply remove the
    # entire prefix
    if user_text.endswith((":", "-", " ")):
        start_index = len(user_text)

    # Look for groups of word chars, replace text until the start of the final group
    else:
        start_indices = [m.start() for m in WORD.finditer(user_text)] or [
            len(user_text)
        ]
        start_index = max(start_indices)

    item = _render_role_common(name, role)
    item.insert_text = insert_text[start_index:]
    return item


def _render_role_with_text_edit(
    context: CompletionContext, name: str, role: Any
) -> Optional[CompletionItem]:
    """Render a role's ``CompletionItem`` using the ``textEdit`` field.

    This implements the ``replace`` insert behavior for roles.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    name
       The name of the role, as it appears in an rst file.

    role
       The implementation of the role.
    """
    match = context.match
    groups = match.groupdict()
    domain = groups["domain"] or ""

    if not name.startswith(domain):
        return None

    # Insert text starting from the starting ':' character of the role.
    start = match.span()[0] + match.group(0).find(":")
    end = start + len(groups["role"])

    range_ = Range(
        start=Position(line=context.position.line, character=start),
        end=Position(line=context.position.line, character=end),
    )

    insert_text = f":{name}:"

    item = _render_role_common(name, role)
    item.filter_text = insert_text
    item.text_edit = TextEdit(range=range_, new_text=insert_text)

    return item


def _render_role_common(name: str, role: Any) -> CompletionItem:
    """Render the common fields of a role's completion item."""

    try:
        dotted_name = f"{role.__module__}.{role.__name__}"
    except AttributeError:
        dotted_name = f"{role.__module__}.{role.__class__.__name__}"

    return CompletionItem(
        label=name,
        kind=CompletionItemKind.Function,
        detail=f"{dotted_name}",
        data={"completion_type": "role"},
    )



import inspect
import logging
import os.path
import pathlib
import sys
import typing
from functools import partial
from typing import IO
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from uuid import uuid4

from sphinx import __version__ as __sphinx_version__
from sphinx.application import Sphinx
from sphinx.util import console
from sphinx.util import logging as sphinx_logging_module
from sphinx.util.logging import NAMESPACE as SPHINX_LOG_NAMESPACE
from sphinx.util.logging import VERBOSITY_MAP

from . import types
from .config import SphinxConfig
from .log import SphinxLogHandler
from .transforms import LineNumberTransform
from .util import send_error
from .util import send_message

STATIC_DIR = (pathlib.Path(__file__).parent / "static").resolve()


class SphinxHandler:
    """Responsible for implementing the JSON-RPC API exposed by the Sphinx agent."""

    def __init__(self):
        self.app: Optional[Sphinx] = None
        """The sphinx application instance"""

        self.log_handler: Optional[SphinxLogHandler] = None
        """The logging handler"""

        self._content_overrides: Dict[str, str] = {}
        """Holds any additional content to inject into a build."""

        self._handlers: Dict[str, Tuple[Type, Callable]] = self._register_handlers()

    def get(self, method: str) -> Optional[Tuple[Type, Callable]]:
        """Return the handler for the given method - if possible.

        Parameters
        ----------
        method
           The name of the method

        Returns
        -------
        Optional[Tuple[Type, Callable]]
           A tuple where the first element is the type definition
           representing the message body, the second element is the method which
           implements it.

           If ``None``, the given method is unknown.

        """
        return self._handlers.get(method)

    def _register_handlers(self) -> Dict[str, Tuple[Type, Callable]]:
        """Return a map of all the handlers we provide.

        A handler

        - must be a method on this class
        - the must take a single parameter called ``request``
        - the type annotation for that parameter must correspond with a ``XXXRequest``
          class definition from the ``types`` module.

        Returns
        -------
        Dict[str, Tuple[Type, Callable]]
           A dictonary where the keys are the method names implemented by this class.
           Values are a tuple where the first element is the type definition
           representing the message body, the second element is the method which
           implements it.
        """
        handlers: Dict[str, Tuple[Type, Callable]] = {}

        for name in dir(self):
            method_func = getattr(self, name)
            if name.startswith("_") or not inspect.ismethod(method_func):
                continue

            parameters = inspect.signature(method_func).parameters
            if set(parameters.keys()) != {"request"}:
                continue

            request_type = typing.get_type_hints(method_func)["request"]
            if not all(
                [hasattr(request_type, "jsonrpc"), hasattr(request_type, "method")]
            ):
                continue

            handlers[request_type.method] = (request_type, method_func)

        return handlers

    def create_sphinx_app(self, request: types.CreateApplicationRequest):
        """Create a new sphinx application instance."""
        sphinx_config = SphinxConfig.fromcli(request.params.command)
        if sphinx_config is None:
            raise ValueError("Invalid build command")

        sphinx_args = sphinx_config.to_application_args()

        # Override Sphinx's logging setup with our own.
        sphinx_logging_module.setup = partial(self.setup_logging, sphinx_config)
        self.app = Sphinx(**sphinx_args)

        # Connect event handlers.
        self.app.connect("env-before-read-docs", self._cb_env_before_read_docs)
        self.app.connect("source-read", self._cb_source_read, priority=0)

        # TODO: Sphinx 7.x has introduced a `include-read` event
        # See: https://github.com/sphinx-doc/sphinx/pull/11657

        if request.params.enable_sync_scrolling:
            _enable_sync_scrolling(self.app)

        response = types.CreateApplicationResponse(
            id=request.id,
            result=types.SphinxInfo(
                id=str(uuid4()),
                version=__sphinx_version__,
                conf_dir=str(self.app.confdir),
                build_dir=str(self.app.outdir),
                builder_name=self.app.builder.name,
                src_dir=str(self.app.srcdir),
            ),
            jsonrpc=request.jsonrpc,
        )
        send_message(response)

    def _cb_env_before_read_docs(self, app: Sphinx, env, docnames: List[str]):
        """Used to add additional documents to the "to build" list."""

        is_building = set(docnames)

        for docname in env.found_docs - is_building:
            filepath = env.doc2path(docname, base=True)
            if filepath in self._content_overrides:
                docnames.append(docname)

    def _cb_source_read(self, app: Sphinx, docname: str, source):
        """Called whenever sphinx reads a file from disk."""

        filepath = app.env.doc2path(docname, base=True)

        # Clear diagnostics
        if self.log_handler:
            self.log_handler.diagnostics.pop(filepath, None)

        # Override file contents if necessary
        if (content := self._content_overrides.get(filepath)) is not None:
            source[0] = content

    def setup_logging(self, config: SphinxConfig, app: Sphinx, status: IO, warning: IO):
        """Setup Sphinx's logging so that it integrates well with the parent language
        server."""

        # Disable color escape codes in Sphinx's log messages
        console.nocolor()

        if not config.silent:
            sphinx_logger = logging.getLogger(SPHINX_LOG_NAMESPACE)

            # Be sure to remove any old handlers
            for handler in sphinx_logger.handlers:
                if isinstance(handler, SphinxLogHandler):
                    sphinx_logger.handlers.remove(handler)
                    self.log_handler = None

            self.log_handler = SphinxLogHandler(app)
            sphinx_logger.addHandler(self.log_handler)

            if config.quiet:
                level = logging.WARNING
            else:
                level = VERBOSITY_MAP[app.verbosity]

            sphinx_logger.setLevel(level)
            self.log_handler.setLevel(level)

            formatter = logging.Formatter("%(message)s")
            self.log_handler.setFormatter(formatter)

    def build_sphinx_app(self, request: types.BuildRequest):
        """Trigger a Sphinx build."""

        if self.app is None:
            send_error(id=request.id, code=-32803, message="Sphinx app not initialized")
            return

        self._content_overrides = request.params.content_overrides

        try:
            self.app.build()

            diagnostics = {}
            if self.log_handler:
                diagnostics = {
                    fpath: list(items)
                    for fpath, items in self.log_handler.diagnostics.items()
                }

            response = types.BuildResponse(
                id=request.id,
                result=types.BuildResult(
                    build_file_map=_build_file_mapping(self.app),
                    diagnostics=diagnostics,
                ),
                jsonrpc=request.jsonrpc,
            )
            send_message(response)
        except Exception:
            send_error(id=request.id, code=-32602, message="Sphinx build failed.")

    def notify_exit(self, request: types.ExitNotification):
        """Sent from the client to signal that the agent should exit."""
        sys.exit(0)


def _build_file_mapping(app: Sphinx) -> Dict[str, str]:
    """Given a Sphinx application, return a mapping of all known source files to their
    corresponding output files."""

    env = app.env
    builder = app.builder
    mapping = {env.doc2path(doc): builder.get_target_uri(doc) for doc in env.found_docs}

    # Don't forget any included files.
    # TODO: How best to handle files included in multiple documents?
    for parent_doc, included_docs in env.included.items():
        for doc in included_docs:
            mapping[env.doc2path(doc)] = mapping[env.doc2path(parent_doc)]

    # Ensure any relative paths in included docs are resolved.
    mapping = {str(pathlib.Path(d).resolve()): uri for d, uri in mapping.items()}

    return mapping


def _enable_sync_scrolling(app: Sphinx):
    """Given a Sphinx application, configure it so that we can support syncronised
    scrolling."""

    # On OSes like Fedora Silverblue where `/home` is a symlink for `/var/home`
    # we could have a situation where `STATIC_DIR` and `app.confdir` have
    # different root dirs... which is enough to cause `os.path.relpath` to return
    # the wrong path.
    #
    # Fully resolving both `STATIC_DIR` and `app.confdir` should be enough to
    # mitigate this.
    confdir = pathlib.Path(app.confdir).resolve()

    # Push our folder of static assets into the user's project.
    # Path needs to be relative to their project's confdir.
    reldir = os.path.relpath(str(STATIC_DIR), start=str(confdir))
    app.config.html_static_path.append(reldir)

    app.add_js_file("webview.js")

    # Inject source line numbers into build output
    app.add_transform(LineNumberTransform)



from typing import List
from typing import Optional

from lsprotocol import types

from esbonio.server import EsbonioLanguageServer
from esbonio.server import Uri
from esbonio.server.feature import LanguageFeature

from .io import read_initial_doctree
from .visitor import SymbolVisitor


class DocumentSymbols(LanguageFeature):
    """Handles document symbol requests."""

    def document_symbol(
        self, params: types.DocumentSymbolParams
    ) -> Optional[List[types.DocumentSymbol]]:
        uri = params.text_document.uri
        doc = self.server.workspace.get_document(uri)

        self.logger.debug("doc: %s %s", doc.language_id, uri)
        if doc.language_id not in {"restructuredtext"}:
            return None

        try:
            self.server.clear_diagnostics("docutils", Uri.parse(doc.uri))
            doctree = read_initial_doctree(doc, self.server.logger)
            self.server.sync_diagnostics()
        except Exception:
            self.logger.error("Unable to parse doctree", exc_info=True)
            return None

        if doctree is None:
            return None

        visitor = SymbolVisitor(self.logger, doctree)
        doctree.walkabout(visitor)

        return visitor.symbols


def esbonio_setup(server: EsbonioLanguageServer):
    document_symbols = DocumentSymbols(server)
    server.add_feature(document_symbols)



import hashlib
import importlib.util
import logging
import pathlib
from typing import List
from typing import Optional

import attrs
import platformdirs
from pygls.workspace import Workspace

from esbonio.server import Uri


def get_module_path(module: str) -> Optional[pathlib.Path]:
    """Return the path to the directory containing the given module name.

    Parameters
    ----------
    module
       A valid Python module name e.g. ``esbonio.sphinx_agent``

    Returns
    -------
    pathlib.Path | None
       The path to the directory containing the given module.
       If ``None``, the module could not be found.
    """
    spec = importlib.util.find_spec(module)
    if spec is None:
        return None

    if spec.origin is None:
        return None

    # origin = .../esbonio/sphinx_agent/__init__.py
    agent = pathlib.Path(spec.origin)
    return agent.parent.parent


@attrs.define
class SphinxConfig:
    """Configuration for the sphinx agent subprocess."""

    enable_dev_tools: bool = attrs.field(default=False)
    """Flag to enable dev tools."""

    enable_sync_scrolling: bool = attrs.field(default=True)
    """Flag to enable sync scrolling."""

    python_command: List[str] = attrs.field(factory=list)
    """The command to use when launching the python interpreter."""

    build_command: List[str] = attrs.field(factory=list)
    """The sphinx-build command to use."""

    env_passthrough: List[str] = attrs.field(factory=list)
    """List of environment variables to pass through to the Sphinx subprocess"""

    cwd: str = attrs.field(default="")
    """The working directory to use."""

    python_path: List[pathlib.Path] = attrs.field(factory=list)
    """The value of ``PYTHONPATH`` to use when injecting the sphinx agent into the
    target environment"""

    def resolve(
        self,
        uri: Uri,
        workspace: Workspace,
        logger: logging.Logger,
    ) -> "Optional[SphinxConfig]":
        """Resolve the configuration based on user provided values.

        Parameters
        ----------
        uri
           The uri of the file we are creating the sphinx agent instace for

        workspace
           The user's workspace

        logger
           The logger instance to use.

        Returns
        -------
        SphinxConfig | None
           The fully resolved config object to use.
           If ``None``, a valid configuration could not be created.
        """
        python_path = self._resolve_python_path(logger)
        if len(python_path) == 0:
            return None

        cwd = self._resolve_cwd(uri, workspace, logger)
        if cwd is None:
            return None

        build_command = self._resolve_build_command(uri, logger)
        if len(build_command) == 0:
            return None

        logger.debug("Cwd: %s", cwd)
        logger.debug("Build command: %s", build_command)

        return SphinxConfig(
            enable_dev_tools=self.enable_dev_tools,
            enable_sync_scrolling=self.enable_sync_scrolling,
            cwd=cwd,
            env_passthrough=self.env_passthrough,
            python_command=self.python_command,
            build_command=build_command,
            python_path=python_path,
        )

    def _resolve_cwd(
        self, uri: Uri, workspace: Workspace, logger: logging.Logger
    ) -> Optional[str]:
        """If no working directory is given, try to determine the appropriate working
        directory based on the workspace.

        Parameters
        ----------
        uri
           The uri of the file we are creating an agent instance for

        workspace
           The user's workspace.

        logger
           The logger instance to use.

        Returns
        -------
        str | None
           The working directory to launch the sphinx agent in.
           If ``None``, the working directory could not be determined.
        """

        if self.cwd:
            return self.cwd

        for folder_uri in workspace.folders.keys():
            folder_uri = Uri.parse(folder_uri)
            if folder_uri and str(uri).startswith(str(folder_uri)):
                break
        else:
            folder_uri = Uri.parse(workspace.root_uri)

        if folder_uri is None or (cwd := folder_uri.fs_path) is None:
            logger.error("Unable to determine working directory from '%s'", folder_uri)
            return None

        return cwd

    def _resolve_python_path(self, logger: logging.Logger) -> List[pathlib.Path]:
        """Return the list of paths to put on the sphinx agent's ``PYTHONPATH``

        Using the ``PYTHONPATH`` environment variable, we can inject additional Python
        packages into the user's Python environment. This method will locate the
        installation path of the sphinx agent and return it.

        Additionally if the ``enable_dev_tools`` flag is set, this will attempt to
        locate the ``lsp_devtools`` package

        Parameters
        ----------
        logger
           The logger instance to use

        Returns
        -------
        List[pathlib.Path]
           The list of paths to Python packages to inject into the sphinx agent's target
           environment. If empty, the ``esbonio.sphinx_agent`` package was not found.
        """
        if len(self.python_path) > 0:
            return self.python_path

        if (sphinx_agent := get_module_path("esbonio.sphinx_agent")) is None:
            logger.error("Unable to locate the sphinx agent")
            return []

        python_path = [sphinx_agent]
        if self.enable_dev_tools:
            if (lsp_devtools := get_module_path("lsp_devtools")) is None:
                logger.warning(
                    "Unable to locate module 'lsp_devtools', dev tools will not "
                    "be availble."
                )
            else:
                python_path.append(lsp_devtools)

        return python_path

    def _resolve_build_command(self, uri: Uri, logger: logging.Logger) -> List[str]:
        """Return the ``sphinx-build`` command to use.

        If no command is configured, this will attempt to guess the command to use based
        on the user's workspace.

        Parameters
        ----------
        uri
           The uri of the file we are creating the sphinx agent for.

        logger
           The logger instance to use.

        Returns
        -------
        List[str]
           The ``sphinx-build`` command to use.
           If empty, no build command could be determined.
        """

        if len(self.build_command) > 0:
            return self.build_command

        if (path := uri.fs_path) is None:
            return []

        # Search upwards from the given uri to see if we find something that looks like
        # a sphinx conf.py file.
        previous = None
        current = pathlib.Path(path)

        while previous != current:
            previous = current
            current = previous.parent

            conf_py = current / "conf.py"
            logger.debug("Trying path: %s", current)
            if conf_py.exists():
                cache = platformdirs.user_cache_dir("esbonio", "swyddfa")
                project = hashlib.md5(str(current).encode()).hexdigest()
                build_dir = str(pathlib.Path(cache, project))
                return ["sphinx-build", "-M", "dirhtml", str(current), str(build_dir)]

        return []



import json
import logging
import pathlib
import platform
import traceback
import typing
import warnings
from functools import partial
from multiprocessing import Process
from multiprocessing import Queue
from typing import IO
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import pygls.uris as Uri
from lsprotocol.types import DeleteFilesParams
from lsprotocol.types import Diagnostic
from lsprotocol.types import DiagnosticSeverity
from lsprotocol.types import DidSaveTextDocumentParams
from lsprotocol.types import InitializedParams
from lsprotocol.types import InitializeParams
from lsprotocol.types import MessageType
from lsprotocol.types import Position
from lsprotocol.types import Range
from lsprotocol.types import ShowDocumentParams
from sphinx import __version__ as __sphinx_version__
from sphinx.application import Sphinx
from sphinx.domains import Domain
from sphinx.errors import ConfigError
from sphinx.util import console
from sphinx.util import logging as sphinx_logging_module
from sphinx.util.logging import NAMESPACE as SPHINX_LOG_NAMESPACE
from sphinx.util.logging import VERBOSITY_MAP

from esbonio.cli import setup_cli
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.sphinx.config import InitializationOptions
from esbonio.lsp.sphinx.config import MissingConfigError
from esbonio.lsp.sphinx.config import SphinxConfig
from esbonio.lsp.sphinx.config import SphinxLogHandler
from esbonio.lsp.sphinx.config import SphinxServerConfig
from esbonio.lsp.sphinx.preview import make_preview_server
from esbonio.lsp.sphinx.preview import start_preview_server

from .line_number_transform import LineNumberTransform

__all__ = [
    "InitializationOptions",
    "MissingConfigError",
    "SphinxConfig",
    "SphinxServerConfig",
    "SphinxLanguageServer",
]

IS_LINUX = platform.system() == "Linux"

# fmt: off
# Order matters!
DEFAULT_MODULES = [
    "esbonio.lsp.directives",         # Generic directive support
    "esbonio.lsp.roles",              # Generic roles support
    "esbonio.lsp.rst.directives",     # docutils directives
    "esbonio.lsp.rst.roles",          # docutils roles
    "esbonio.lsp.sphinx.autodoc",     # automodule, autoclass, etc.
    "esbonio.lsp.sphinx.codeblocks",  # code-block, highlight, etc.
    "esbonio.lsp.sphinx.domains",     # Sphinx domains
    "esbonio.lsp.sphinx.directives",  # Sphinx directives
    "esbonio.lsp.sphinx.images",      # image, figure etc
    "esbonio.lsp.sphinx.includes",    # include, literal-include etc.
    "esbonio.lsp.sphinx.roles",       # misc roles added by Sphinx e.g. :download:
]
"""The modules to load in the default configuration of the server."""
# fmt: on


class SphinxLanguageServer(RstLanguageServer):
    """A language server dedicated to working with Sphinx projects."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app: Optional[Sphinx] = None
        """The Sphinx application instance."""

        self.sphinx_args: Dict[str, Any] = {}
        """The current Sphinx configuration will all variables expanded."""

        self.sphinx_log: Optional[SphinxLogHandler] = None
        """Logging handler for sphinx messages."""

        self.preview_process: Optional[Process] = None
        """The process hosting the preview server."""

        self.preview_port: Optional[int] = None
        """The port the preview server is running on."""

        self._role_target_types: Optional[Dict[str, List[str]]] = None
        """Cache for role target types."""

        self.file_list_pending_build_version_updates: List[Tuple[str, int]] = []
        """List of all the files that need an updated last_build_version"""

    @property
    def configuration(self) -> Dict[str, Any]:
        """Return the server's actual configuration."""
        config = super().configuration
        sphinx_config = SphinxConfig.from_arguments(sphinx_args=self.sphinx_args)

        if sphinx_config is None:
            self.logger.error("Unable to determine SphinxConfig!")
            return config

        if self.user_config is None:
            self.logger.error("Unable to determine user config!")
            return config

        # We always run Sphinx in "'-Q' mode", so we need to go back to the user's
        # config to get those values.
        sphinx_config.silent = self.user_config.sphinx.silent  # type: ignore
        sphinx_config.quiet = self.user_config.sphinx.quiet  # type: ignore

        # 'Make mode' isn't something that can be inferred from Sphinx args either.
        sphinx_config.make_mode = self.user_config.sphinx.make_mode  # type: ignore

        config["sphinx"] = self.converter.unstructure(sphinx_config)
        config["sphinx"]["command"] = ["sphinx-build"] + sphinx_config.to_cli_args()
        config["sphinx"]["version"] = __sphinx_version__

        config["server"] = self.converter.unstructure(self.user_config.server)

        return config

    def initialize(self, params: InitializeParams):
        super().initialize(params)
        self.user_config = self.converter.structure(
            params.initialization_options or {}, InitializationOptions
        )

    def initialized(self, params: InitializedParams):
        self.app = self._initialize_sphinx()
        self.build()

    def _initialize_sphinx(self):
        try:
            return self.create_sphinx_app(self.user_config)  # type: ignore
        except MissingConfigError:
            self.show_message(
                message="Unable to find your 'conf.py', features that depend on Sphinx will be unavailable",
                msg_type=MessageType.Warning,
            )
            self.send_notification(
                "esbonio/buildComplete",
                {
                    "config": self.configuration,
                    "error": True,
                    "warnings": 0,
                },
            )
        except Exception as exc:
            self.logger.error(traceback.format_exc())
            uri, diagnostic = exception_to_diagnostic(exc)
            self.set_diagnostics("conf.py", uri, [diagnostic])

            self.sync_diagnostics()
            self.send_notification(
                "esbonio/buildComplete",
                {"config": self.configuration, "error": True, "warnings": 0},
            )

    def on_shutdown(self, *args):
        if self.preview_process:
            if not hasattr(self.preview_process, "kill"):
                self.preview_process.terminate()
            else:
                self.preview_process.kill()

    def save(self, params: DidSaveTextDocumentParams):
        super().save(params)

        filepath = Uri.to_fs_path(params.text_document.uri)
        if filepath.endswith("conf.py"):
            if self.app:
                conf_dir = pathlib.Path(self.app.confdir)
            else:
                # The user's config is currently broken... where should their conf.py be?
                if self.user_config is not None:
                    config = typing.cast(InitializationOptions, self.user_config).sphinx
                else:
                    config = SphinxConfig()

                conf_dir = config.resolve_conf_dir(
                    self.workspace.root_uri
                ) or pathlib.Path(".")

            if str(conf_dir / "conf.py") == filepath:
                self.clear_diagnostics("conf.py")
                self.sync_diagnostics()
                self.app = self._initialize_sphinx()

        self.build()

    def delete_files(self, params: DeleteFilesParams):
        self.logger.debug("Deleted files: %s", params.files)

        # Files don't exist anymore, so diagnostics must be cleared.
        for file in params.files:
            self.clear_diagnostics("sphinx", file.uri)

        self.build()

    def build(
        self, force_all: bool = False, filenames: Optional[List[str]] = None
    ) -> None:
        """Trigger sphinx build. Force complete rebuild with flag or build only selected files in the list."""
        if not self.app:
            return

        self.logger.debug("Building...")
        self.send_notification("esbonio/buildStart", {})
        self.clear_diagnostics("sphinx-build")
        self.sync_diagnostics()

        # Reset the warnings counter
        self.app._warncount = 0
        error = False

        if self.sphinx_log is not None:
            self.sphinx_log.diagnostics = {}

        try:
            self.app.build(force_all, filenames)
        except Exception as exc:
            error = True
            self.logger.error(traceback.format_exc())
            uri, diagnostic = exception_to_diagnostic(exc)
            self.set_diagnostics("sphinx-build", uri, [diagnostic])

        if self.sphinx_log is not None:
            for doc, diagnostics in self.sphinx_log.diagnostics.items():
                self.logger.debug("Found %d problems for %s", len(diagnostics), doc)
                self.set_diagnostics("sphinx", doc, diagnostics)

        self.sync_diagnostics()

        self.send_notification(
            "esbonio/buildComplete",
            {
                "config": self.configuration,
                "error": error,
                "warnings": self.app._warncount,
            },
        )

    def cb_env_before_read_docs(self, app, env, docnames: List[str]):
        """Callback handling env-before-read-docs event."""

        # Determine if any unsaved files need to be added to the build list
        if self.user_config.server.enable_live_preview:  # type: ignore
            is_building = set(docnames)

            for docname in env.found_docs - is_building:
                filepath = env.doc2path(docname, base=True)
                uri = Uri.from_fs_path(filepath)

                doc = self.workspace.get_document(uri)
                current_version = doc.version or 0

                last_build_version = getattr(doc, "last_build_version", 0)
                if last_build_version < current_version:
                    docnames.append(docname)

        # Clear diagnostics for any to-be built files
        for docname in docnames:
            filepath = env.doc2path(docname, base=True)
            uri = Uri.from_fs_path(filepath)
            self.clear_diagnostics("sphinx", uri)

            doc = self.workspace.get_document(uri)
            current_version = doc.version or 0
            self.file_list_pending_build_version_updates.append((uri, current_version))  # type: ignore

    def cb_build_finished(self, app, exception):
        """Callback handling build-finished event."""
        if exception:
            self.file_list_pending_build_version_updates = []
            return

        for uri, updated_version in self.file_list_pending_build_version_updates:
            doc = self.workspace.get_document(uri)
            last_build_version = getattr(doc, "last_build_version", 0)
            if last_build_version < updated_version:
                doc.last_build_version = updated_version  # type: ignore

        self.file_list_pending_build_version_updates = []

    def cb_source_read(self, app, docname, source):
        """Callback handling source_read event."""

        if not self.user_config.server.enable_live_preview:  # type: ignore
            return

        filepath = app.env.doc2path(docname, base=True)
        uri = Uri.from_fs_path(filepath)

        doc = self.workspace.get_document(uri)
        source[0] = doc.source

    def create_sphinx_app(self, options: InitializationOptions) -> Optional[Sphinx]:
        """Create a Sphinx application instance with the given config."""
        sphinx = options.sphinx
        server = options.server
        self.logger.debug(
            "User Config %s", json.dumps(self.converter.unstructure(sphinx), indent=2)
        )

        # Until true multi-root support can be implemented let's try each workspace
        # folder and use the first valid configuration we can find.
        for folder_uri in self.workspace.folders.keys():
            self.logger.debug("Workspace Folder: '%s'", folder_uri)

            try:
                sphinx_config = sphinx.resolve(folder_uri)
                break
            except MissingConfigError:
                self.logger.debug(
                    "No Sphinx conifg found in workspace folder: '%s'", folder_uri
                )

        # Not all clients use/support workspace folders, as a fallback, try the root_uri.
        else:
            self.logger.debug("Workspace root '%s'", self.workspace.root_uri)
            sphinx_config = sphinx.resolve(self.workspace.root_uri)

        self.sphinx_args = sphinx_config.to_application_args()
        self.logger.debug("Sphinx Args %s", json.dumps(self.sphinx_args, indent=2))

        # Override Sphinx's logging setup with our own.
        sphinx_logging_module.setup = partial(self._logging_setup, server, sphinx)
        app = Sphinx(**self.sphinx_args)

        self._load_sphinx_extensions(app)
        self._load_sphinx_config(app)

        if self.user_config.server.enable_scroll_sync:  # type: ignore
            app.add_transform(LineNumberTransform)

        app.connect("env-before-read-docs", self.cb_env_before_read_docs)

        if self.user_config.server.enable_live_preview:  # type: ignore
            app.connect("source-read", self.cb_source_read, priority=0)
            app.connect("build-finished", self.cb_build_finished)

        return app

    def _logging_setup(
        self,
        server: SphinxServerConfig,
        sphinx: SphinxConfig,
        app: Sphinx,
        status: IO,
        warning: IO,
    ):
        # Disable color escape codes in Sphinx's log messages
        console.nocolor()

        if not server.hide_sphinx_output and not sphinx.silent:
            sphinx_logger = logging.getLogger(SPHINX_LOG_NAMESPACE)

            # Be sure to remove any old handlers.
            for handler in sphinx_logger.handlers:
                if isinstance(handler, SphinxLogHandler):
                    sphinx_logger.handlers.remove(handler)

            self.sphinx_log = SphinxLogHandler(app, self)
            sphinx_logger.addHandler(self.sphinx_log)

            if sphinx.quiet:
                level = logging.WARNING
            else:
                level = VERBOSITY_MAP[app.verbosity]

            sphinx_logger.setLevel(level)
            self.sphinx_log.setLevel(level)

            formatter = logging.Formatter("%(message)s")
            self.sphinx_log.setFormatter(formatter)

    def _load_sphinx_extensions(self, app: Sphinx):
        """Loop through each of Sphinx's extensions and see if any contain server
        functionality.
        """

        for name, ext in app.extensions.items():
            mod = ext.module

            setup = getattr(mod, "esbonio_setup", None)
            if setup is None:
                self.logger.debug(
                    "Skipping extension '%s', missing 'esbonio_setup' fuction", name
                )
                continue

            self.load_extension(name, setup)

    def _load_sphinx_config(self, app: Sphinx):
        """Try and load the config as an server extension."""

        name = "<sphinx-config>"

        setup = app.config._raw_config.get("esbonio_setup", None)
        if not setup or not callable(setup):
            return

        self.load_extension(name, setup)

    def preview(self, options: Dict[str, Any]) -> Dict[str, Any]:
        if not self.app or not self.app.builder:
            return {}

        builder_name = self.app.builder.name
        if builder_name not in {"html"}:
            self.show_message(
                f"Previews are not currently supported for the '{builder_name}' builder."
            )

            return {}

        if not self.preview_process and IS_LINUX:
            self.logger.debug("Starting preview server.")
            server = make_preview_server(self.app.outdir)  # type: ignore[arg-type]
            self.preview_port = server.server_port

            self.preview_process = Process(target=server.serve_forever, daemon=True)
            self.preview_process.start()

        if not self.preview_process and not IS_LINUX:
            self.logger.debug("Starting preview server")

            q: Queue = Queue()
            self.preview_process = Process(
                target=start_preview_server, args=(q, self.app.outdir), daemon=True
            )
            self.preview_process.start()
            self.preview_port = q.get()

        if options.get("show", True):
            self.show_document(
                ShowDocumentParams(
                    uri=f"http://localhost:{self.preview_port}", external=True
                )
            )

        return {"port": self.preview_port}

    def get_doctree(
        self, *, docname: Optional[str] = None, uri: Optional[str] = None
    ) -> Optional[Any]:
        """Return the initial doctree corresponding to the specified document.

        The ``docname`` of a document is its path relative to the project's ``srcdir``
        minus the extension e.g. the docname of the file ``docs/lsp/features.rst``
        would be ``lsp/features``.

        Parameters
        ----------
        docname:
           Returns the doctree that corresponds with the given docname
        uri:
           Returns the doctree that corresponds with the given uri.
        """

        if self.app is None or self.app.env is None or self.app.builder is None:
            return None

        if uri is not None:
            fspath = Uri.to_fs_path(uri)
            docname = self.app.env.path2doc(fspath)

        if docname is None:
            return None

        try:
            return self.app.env.get_and_resolve_doctree(docname, self.app.builder)
        except FileNotFoundError:
            self.logger.debug("Could not find doctree for '%s'", docname)
            # self.logger.debug(traceback.format_exc())
            return None

    def get_domain(self, name: str) -> Optional[Domain]:
        """Return the domain with the given name.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0``

        If a domain with the given name cannot be found, this method will return None.

        Parameters
        ----------
        name:
           The name of the domain
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_domains() is deprecated and will be removed in v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        if self.app is None or self.app.env is None:
            return None

        domains = self.app.env.domains
        return domains.get(name, None)

    def get_domains(self) -> Iterator[Tuple[str, Domain]]:
        """Get all the domains registered with an applications.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0``

        Returns a generator that iterates through all of an application's domains,
        taking into account configuration variables such as ``primary_domain``.
        Yielded values will be a tuple of the form ``(prefix, domain)`` where

        - ``prefix`` is the namespace that should be used when referencing items
          in the domain
        - ``domain`` is the domain object itself.

        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_domains() is deprecated and will be removed in v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        if self.app is None or self.app.env is None:
            return []

        domains = self.app.env.domains
        primary_domain = self.app.config.primary_domain

        for name, domain in domains.items():
            prefix = name

            # Items from the standard and primary domains don't require the namespace prefix
            if name == "std" or name == primary_domain:
                prefix = ""

            yield prefix, domain

    def get_directive_options(self, name: str) -> Dict[str, Any]:
        """Return the options specification for the given directive.

        .. deprecated:: 0.14.2

           This will be removed in ``v1.0``
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_directive_options() is deprecated and will be removed in "
            "v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        directive = self.get_directives().get(name, None)
        if directive is None:
            return {}

        options = directive.option_spec

        if name.startswith("auto") and self.app:
            self.logger.debug("Processing options for '%s' directive", name)
            name = name.replace("auto", "")

            self.logger.debug("Documenter name is '%s'", name)
            documenter = self.app.registry.documenters.get(name, None)

            if documenter is not None:
                options = documenter.option_spec

        return options or {}

    def get_default_role(self) -> Tuple[Optional[str], Optional[str]]:
        """Return the project's default role"""

        if not self.app:
            return None, None

        role = self.app.config.default_role
        if not role:
            return None, None

        if ":" in role:
            domain, name = role.split(":")

            if domain == self.app.config.primary_domain:
                domain = ""

            return domain, name

        return None, role

    def get_role_target_types(self, name: str, domain_name: str = "") -> List[str]:
        """Return a map indicating which object types a role is capable of linking
        with.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0``

        For example

        .. code-block:: python

           {
               "func": ["function"],
               "class": ["class", "exception"]
           }
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_role_target_types() is deprecated and will be removed in "
            "v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        key = f"{domain_name}:{name}" if domain_name else name

        if self._role_target_types is not None:
            return self._role_target_types.get(key, [])

        self._role_target_types = {}

        for prefix, domain in self.get_domains():
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for name, item_type in domain.object_types.items():
                for role in item_type.roles:
                    role_key = fmt.format(name=role, prefix=prefix)
                    target_types = self._role_target_types.get(role_key, list())
                    target_types.append(name)

                    self._role_target_types[role_key] = target_types

        types = self._role_target_types.get(key, [])
        self.logger.debug("Role '%s' targets object types '%s'", key, types)

        return types

    def get_role_targets(self, name: str, domain: str = "") -> List[tuple]:
        """Return a list of objects targeted by the given role.

        Parameters
        ----------
        name:
           The name of the role
        domain:
           The domain the role is a part of, if applicable.
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_role_targets() is deprecated and will be removed in "
            "v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        targets: List[tuple] = []
        domain_obj: Optional[Domain] = None

        if domain:
            domain_obj = self.get_domain(domain)
        else:
            std = self.get_domain("std")
            if std and name in std.roles:
                domain_obj = std

            elif self.app and self.app.config.primary_domain:
                domain_obj = self.get_domain(self.app.config.primary_domain)

        target_types = set(self.get_role_target_types(name, domain))

        if not domain_obj:
            self.logger.debug("Unable to find domain for role '%s:%s'", domain, name)
            return []

        for obj in domain_obj.get_objects():
            if obj[2] in target_types:
                targets.append(obj)

        return targets

    def get_intersphinx_projects(self) -> List[str]:
        """Return the list of configured intersphinx project names.

        .. deprecated:: 0.15.0

           This will be removed in ``v.1.0``
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_intersphinx_projects() is deprecated and will be removed in "
            "v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        if self.app is None:
            return []

        inv = getattr(self.app.env, "intersphinx_named_inventory", {})
        return list(inv.keys())

    def has_intersphinx_targets(
        self, project: str, name: str, domain: str = ""
    ) -> bool:
        """Return ``True`` if the given intersphinx project has targets targeted by the
        given role.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0``

        Parameters
        ----------
        project:
           The project to check
        name:
           The name of the role
        domain:
           The domain the role is a part of, if applicable.
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.has_intersphinx_targets() is deprecated and will be removed in "
            "v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        targets = self.get_intersphinx_targets(project, name, domain)

        if len(targets) == 0:
            return False

        return any([len(items) > 0 for items in targets.values()])

    def get_intersphinx_targets(
        self, project: str, name: str, domain: str = ""
    ) -> Dict[str, Dict[str, tuple]]:
        """Return the intersphinx objects targeted by the given role.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0``

        Parameters
        ----------
        project:
           The project to return targets from
        name:
           The name of the role
        domain:
           The domain the role is a part of, if applicable.
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_intersphinx_targets() is deprecated and will be removed in "
            "v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        if self.app is None:
            return {}

        inv = getattr(self.app.env, "intersphinx_named_inventory", {})
        if project not in inv:
            return {}

        targets = {}
        inv = inv[project]

        for target_type in self.get_role_target_types(name, domain):
            explicit_domain = f"{domain}:{target_type}"
            if explicit_domain in inv:
                targets[target_type] = inv[explicit_domain]
                continue

            primary_domain = f'{self.app.config.primary_domain or ""}:{target_type}'
            if primary_domain in inv:
                targets[target_type] = inv[primary_domain]
                continue

            std_domain = f"std:{target_type}"
            if std_domain in inv:
                targets[target_type] = inv[std_domain]

        return targets


def exception_to_diagnostic(exc: BaseException):
    """Convert an exception into a diagnostic we can send to the client."""

    # Config errors sometimes wrap the true cause of the problem
    if isinstance(exc, ConfigError) and exc.__cause__ is not None:
        exc = exc.__cause__

    if isinstance(exc, SyntaxError):
        path = pathlib.Path(exc.filename or "")
        line = (exc.lineno or 1) - 1
    else:
        tb = exc.__traceback__
        frame = traceback.extract_tb(tb)[-1]
        path = pathlib.Path(frame.filename)
        line = (frame.lineno or 1) - 1

    message = type(exc).__name__ if exc.args.count == 0 else exc.args[0]

    diagnostic = Diagnostic(
        range=Range(
            start=Position(line=line, character=0),
            end=Position(line=line + 1, character=0),
        ),
        message=message,
        severity=DiagnosticSeverity.Error,
    )

    return Uri.from_fs_path(str(path)), diagnostic


cli = setup_cli("esbonio.lsp.sphinx", "Esbonio's Sphinx language server.")
cli.set_defaults(modules=DEFAULT_MODULES)
cli.set_defaults(server_cls=SphinxLanguageServer)



import json

from esbonio.lsp.directives import Directives
from esbonio.lsp.sphinx import SphinxLanguageServer
from esbonio.lsp.util import resources


def esbonio_setup(rst: SphinxLanguageServer, directives: Directives):
    sphinx_docs = resources.read_string("esbonio.lsp.sphinx", "directives.json")
    directives.add_documentation(json.loads(sphinx_docs))



from typing import List

import attrs

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore[assignment]


@attrs.define
class ServerCompletionConfig:
    """Configuration options for the server that control completion behavior."""

    preferred_insert_behavior: Literal["insert", "replace"] = attrs.field(
        default="replace"
    )
    """This option indicates if the user prefers we use ``insertText`` or ``textEdit``
    when rendering ``CompletionItems``."""


@attrs.define
class ServerConfig:
    """Configuration options for the server."""

    completion: ServerCompletionConfig = attrs.field(factory=ServerCompletionConfig)
    """Configuration values that affect completion"""

    enable_scroll_sync: bool = attrs.field(default=False)
    """Enable custom transformation to add classes with line numbers"""

    enable_live_preview: bool = attrs.field(default=False)
    """Set it to True if you want to build Sphinx app on change event"""

    log_filter: List[str] = attrs.field(factory=list)
    """A list of logger names to restrict output to."""

    log_level: str = attrs.field(default="error")
    """The logging level of server messages to display."""

    show_deprecation_warnings: bool = attrs.field(default=False)
    """Developer flag to enable deprecation warnings."""


@attrs.define
class InitializationOptions:
    """The initialization options we can expect to receive from a client."""

    server: ServerConfig = attrs.field(factory=ServerConfig)
    """The ``esbonio.server.*`` namespace of options."""



"""Type definitions for the sphinx agent.

This is the *only* file shared between the agent itself and the parent language server.
For this reason this file *cannot* import anything from Sphinx.
"""
import dataclasses
import enum
from typing import Dict
from typing import List
from typing import Union


@dataclasses.dataclass(frozen=True)
class Position:
    line: int
    character: int


@dataclasses.dataclass(frozen=True)
class Range:
    start: Position
    end: Position


class DiagnosticSeverity(enum.IntEnum):
    Error = 1
    Warning = 2
    Information = 3
    Hint = 4


@dataclasses.dataclass(frozen=True)
class Diagnostic:
    range: Range
    message: str
    severity: DiagnosticSeverity


@dataclasses.dataclass
class CreateApplicationParams:
    """Parameters of a ``sphinx/createApp`` request."""

    command: List[str]
    """The ``sphinx-build`` command to base the app instance on."""

    enable_sync_scrolling: bool
    """Enable/disable sync scolling of previews."""


@dataclasses.dataclass
class CreateApplicationRequest:
    """A ``sphinx/createApp`` request."""

    id: Union[int, str]

    params: CreateApplicationParams

    method: str = "sphinx/createApp"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class SphinxInfo:
    """Represents information about an instance of the Sphinx application."""

    id: str
    """A unique id representing a particular Sphinx application instance."""

    version: str
    """The version of Sphinx being used."""

    conf_dir: str
    """The folder containing the project's conf.py"""

    build_dir: str
    """The folder containing the Sphinx application's build output"""

    builder_name: str
    """The name of the builder in use"""

    src_dir: str
    """The folder containing the source files for the project"""


@dataclasses.dataclass
class CreateApplicationResponse:
    """A ``sphinx/createApp`` response."""

    id: Union[int, str]

    result: SphinxInfo

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class BuildParams:
    """Parameters of a ``sphinx/build`` request."""

    filenames: List[str] = dataclasses.field(default_factory=list)

    force_all: bool = False

    content_overrides: Dict[str, str] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class BuildResult:
    """Results from a ``sphinx/build`` request."""

    diagnostics: Dict[str, List[Diagnostic]] = dataclasses.field(default_factory=dict)
    """Any diagnostics associated with the project."""

    build_file_map: Dict[str, str] = dataclasses.field(default_factory=dict)
    """A mapping of source files to the output files they contributed to."""


@dataclasses.dataclass
class BuildRequest:
    """A ``sphinx/build`` request."""

    id: Union[int, str]

    params: BuildParams

    method: str = "sphinx/build"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class BuildResponse:
    """A ``sphinx/build`` response."""

    id: Union[int, str]

    result: BuildResult

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class LogMessageParams:
    """Parameters of a ``window/logMessage`` notification."""

    type: int

    message: str


@dataclasses.dataclass
class LogMessage:
    """A ``window/logMessage`` notification"""

    params: LogMessageParams

    method: str = "window/logMessage"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class ExitNotification:
    """An ``exit`` notification"""

    params: None

    method: str = "exit"

    jsonrpc: str = dataclasses.field(default="2.0")


METHOD_TO_MESSAGE_TYPE = {
    BuildRequest.method: BuildRequest,
    ExitNotification.method: ExitNotification,
    CreateApplicationRequest.method: CreateApplicationRequest,
}
METHOD_TO_RESPONSE_TYPE = {
    BuildRequest.method: BuildResponse,
    ExitNotification.method: None,
    CreateApplicationRequest.method: CreateApplicationResponse,
}



from __future__ import annotations

import collections
import inspect
import json
import logging
import typing
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar

import attrs
from lsprotocol import types
from pygls.server import LanguageServer
from pygls.workspace import Document
from pygls.workspace import Workspace

from ._uri import Uri
from .log import setup_logging

if typing.TYPE_CHECKING:
    from .feature import LanguageFeature

__version__ = "0.16.1"
T = TypeVar("T")
LF = TypeVar("LF", bound="LanguageFeature")


@attrs.define
class ServerConfig:
    """Configuration options for the server."""

    log_filter: List[str] = attrs.field(factory=list)
    """A list of logger names to restrict output to."""

    log_level: str = attrs.field(default="error")
    """The logging level of server messages to display."""

    show_deprecation_warnings: bool = attrs.field(default=False)
    """Developer flag to enable deprecation warnings."""


class EsbonioWorkspace(Workspace):
    """A modified version of pygls' workspace that ensures uris are always resolved."""

    def get_document(self, doc_uri: str) -> Document:
        uri = str(Uri.parse(doc_uri).resolve())
        return super().get_document(uri)

    def put_document(self, text_document: types.TextDocumentItem):
        text_document.uri = str(Uri.parse(text_document.uri).resolve())
        return super().put_document(text_document)

    def remove_document(self, doc_uri: str):
        doc_uri = str(Uri.parse(doc_uri).resolve())
        return super().remove_document(doc_uri)

    def update_document(
        self,
        text_doc: types.VersionedTextDocumentIdentifier,
        change: types.TextDocumentContentChangeEvent,
    ):
        text_doc.uri = str(Uri.parse(text_doc.uri).resolve())
        return super().update_document(text_doc, change)


class EsbonioLanguageServer(LanguageServer):
    """The Esbonio language server"""

    def __init__(self, logger: Optional[logging.Logger] = None, *args, **kwargs):
        if "name" not in kwargs:
            kwargs["name"] = "esbonio"

        if "version" not in kwargs:
            kwargs["version"] = __version__

        super().__init__(*args, **kwargs)

        self._diagnostics: Dict[Tuple[str, Uri], List[types.Diagnostic]] = {}
        """Where we store and manage diagnostics."""

        self._loaded_extensions: Dict[str, Any] = {}
        """Record of server modules that have been loaded."""

        self._features: Dict[Type[LanguageFeature], LanguageFeature] = {}
        """The collection of language features registered with the server."""

        self.logger = logger or logging.getLogger(__name__)
        """The logger instance to use."""

        self.converter = self.lsp._converter
        """The cattrs converter instance we should use."""

        self.initialization_options: Optional[types.LSPAny] = None
        """The received initializaion options (if any)"""

    def __iter__(self):
        return iter(self._features.items())

    def initialize(self, params: types.InitializeParams):
        self.logger.info("Initialising esbonio v%s", __version__)
        if (client := params.client_info) is not None:
            self.logger.info("Language client: %s %s", client.name, client.version)

        # TODO: Propose patch to pygls for providing custom Workspace implementations.
        self.lsp._workspace = EsbonioWorkspace(
            self.workspace.root_uri,
            self.workspace._sync_kind,
            list(self.workspace.folders.values()),
        )

        self.initialization_options = params.initialization_options

        # TODO: Merge this with self.get_user_config somehow...
        server_config = ServerConfig()
        if self.initialization_options is not None:
            try:
                config = self.initialization_options.get("server", {})
                server_config = self.converter.structure(config, ServerConfig)
            except Exception:
                self.logger.error("Unable to parse server config", exc_info=True)

        setup_logging(self, server_config)

    def load_extension(self, name: str, setup: Callable):
        """Load the given setup function as an extension.

        If an extension with the given ``name`` already exists, the given setup function
        will be ignored.

        The ``setup`` function can declare dependencies in the form of type
        annotations.

        .. code-block:: python

           from esbonio.lsp.roles import Roles
           from esbonio.lsp.sphinx import SphinxLanguageServer

           def esbonio_setup(rst: SphinxLanguageServer, roles: Roles):
               ...

        In this example the setup function is requesting instances of the
        :class:`~esbonio.lsp.sphinx.SphinxLanguageServer` and the
        :class:`~esbonio.lsp.roles.Roles` language feature.

        Parameters
        ----------
        name
           The name to give the extension

        setup
           The setup function to call
        """

        if name in self._loaded_extensions:
            self.logger.debug("Skipping extension '%s', already loaded", name)
            return

        arguments = _get_setup_arguments(self, setup, name)
        if not arguments:
            return

        try:
            setup(**arguments)

            self.logger.debug("Loaded extension '%s'", name)
            self._loaded_extensions[name] = setup
        except Exception:
            self.logger.error("Unable to load extension '%s'", name, exc_info=True)

    def add_feature(self, feature: LanguageFeature):
        """Register a language feature with the server.

        Parameters
        ----------
        feature
           The language feature
        """
        feature_cls = type(feature)
        if feature_cls in self._features:
            name = f"{feature_cls.__module__}.{feature_cls.__name__}"
            raise RuntimeError(f"Feature '{name}' has already been registered")

        self._features[feature_cls] = feature

    def get_feature(self, feature_cls: Type[LF]) -> Optional[LF]:
        """Returns the requested language feature if it exists, otherwise it returns
        ``None``.

        Parameters
        ----------
        feature_cls
           The class definiion of the feature to retrieve
        """
        return self._features.get(feature_cls, None)  # type: ignore

    async def get_user_config(
        self,
        section: str,
        spec: Type[T],
        scope: Optional[Uri] = None,
    ) -> Optional[T]:
        """Return the user's configuration for the given ``section``.

        Using a ``workspace/configuration`` request, ask the client for the user's
        configuration for the given ``section``.

        ``spec`` should be a class definition representing the expected "shape" of the
        result.

        Parameters
        ----------
        section
           The name of the configuration section to retrieve

        spec
           The class definition representing the expected result.

        scope
           An optional URI, useful in a multi-root context to select which root the
           configuration should be retrieved from.

        Returns
        -------
        T | None
           The user's configuration, parsed as an instance of ``T``.
           If ``None``, the config was not available / there was an error.
        """
        params = types.ConfigurationParams(
            items=[
                types.ConfigurationItem(
                    section=section, scope_uri=str(scope) if scope else None
                )
            ]
        )
        self.logger.debug(
            "workspace/configuration: %s",
            json.dumps(self.converter.unstructure(params), indent=2),
        )
        result = await self.get_configuration_async(params)

        try:
            self.logger.debug("configuration: %s", json.dumps(result[0], indent=2))
            return self.converter.structure(result[0], spec)
        except Exception:
            self.logger.error(
                "Unable to parse configuration as '%s'", spec.__name__, exc_info=True
            )
            return None

    def clear_diagnostics(self, source: str, uri: Optional[Uri] = None) -> None:
        """Clear diagnostics from the given source.

        Parameters
        ----------
        source:
           The source from which to clear diagnostics.
        uri:
           If given, clear diagnostics from within just this uri. Otherwise, all
           diagnostics from the given source are cleared.
        """

        for key in self._diagnostics.keys():
            clear_source = source == key[0]
            clear_uri = uri == key[1] or uri is None

            if clear_source and clear_uri:
                self._diagnostics[key] = []

    def add_diagnostics(self, source: str, uri: Uri, diagnostic: types.Diagnostic):
        """Add a diagnostic to the given source and uri.

        Parameters
        ----------
        source
           The source the diagnostics are from
        uri
           The uri the diagnostics are associated with
        diagnostic
           The diagnostic to add
        """
        key = (source, uri)
        self._diagnostics.setdefault(key, []).append(diagnostic)

    def set_diagnostics(
        self, source: str, uri: Uri, diagnostics: List[types.Diagnostic]
    ) -> None:
        """Set the diagnostics for the given source and uri.

        Parameters
        ----------
        source:
           The source the diagnostics are from
        uri:
           The uri the diagnostics are associated with
        diagnostics:
           The diagnostics themselves
        """
        self._diagnostics[(source, uri)] = diagnostics

    def sync_diagnostics(self) -> None:
        """Update the client with the currently stored diagnostics."""

        uris = {uri for _, uri in self._diagnostics.keys()}
        diagnostics = {uri: DiagnosticList() for uri in uris}

        for (source, uri), diags in self._diagnostics.items():
            for diag in diags:
                diag.source = source
                diagnostics[uri].append(diag)

        for uri, diag_list in diagnostics.items():
            self.logger.debug("Publishing %d diagnostics for: %s", len(diag_list), uri)
            self.publish_diagnostics(str(uri), diag_list.data)


class DiagnosticList(collections.UserList):
    """A list type dedicated to holding diagnostics.

    This is mainly to ensure that only one instance of a diagnostic ever gets
    reported.
    """

    def append(self, item: types.Diagnostic):
        if not isinstance(item, types.Diagnostic):
            raise TypeError("Expected Diagnostic")

        for existing in self.data:
            fields = [
                existing.range == item.range,
                existing.message == item.message,
                existing.severity == item.severity,
                existing.code == item.code,
                existing.source == item.source,
            ]

            if all(fields):
                # Item already added, nothing to do.
                return

        self.data.append(item)


def _get_setup_arguments(
    server: EsbonioLanguageServer, setup: Callable, modname: str
) -> Optional[Dict[str, Any]]:
    """Given a setup function, try to construct the collection of arguments to pass to
    it.
    """
    annotations = typing.get_type_hints(setup)
    parameters = {
        p.name: annotations[p.name]
        for p in inspect.signature(setup).parameters.values()
    }

    args = {}
    for name, type_ in parameters.items():
        if issubclass(server.__class__, type_):
            args[name] = server
            continue

        from .feature import LanguageFeature  # noqa: F402

        if issubclass(type_, LanguageFeature):
            # Try and obtain an instance of the requested language feature.
            feature = server.get_feature(type_)
            if feature is not None:
                args[name] = feature
                continue

            server.logger.debug(
                "Skipping extension '%s', server missing requested feature: '%s'",
                modname,
                type_,
            )
            return None

        server.logger.error(
            "Skipping extension '%s', parameter '%s' has unsupported type: '%s'",
            modname,
            name,
            type_,
        )
        return None

    return args



from __future__ import annotations

from esbonio.server import EsbonioLanguageServer

from .client import SphinxClient
from .client_subprocess import make_subprocess_sphinx_client
from .config import SphinxConfig
from .manager import SphinxManager

__all__ = ["SphinxClient", "SphinxConfig", "SphinxManager"]


def esbonio_setup(server: EsbonioLanguageServer):
    manager = SphinxManager(make_subprocess_sphinx_client, server)
    server.add_feature(manager)



"""Support for Sphinx domains."""
import pathlib
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type

import pygls.uris as Uri
from docutils import nodes
from docutils.parsers.rst import Directive
from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionItemKind
from lsprotocol.types import Location
from lsprotocol.types import Position
from lsprotocol.types import Range
from pygls.workspace import Document
from sphinx.domains import Domain

from esbonio.lsp import CompletionContext
from esbonio.lsp import DefinitionContext
from esbonio.lsp import DocumentLinkContext
from esbonio.lsp.directives import DirectiveLanguageFeature
from esbonio.lsp.directives import Directives
from esbonio.lsp.roles import RoleLanguageFeature
from esbonio.lsp.roles import Roles
from esbonio.lsp.sphinx import SphinxLanguageServer

TARGET_KINDS = {
    "attribute": CompletionItemKind.Field,
    "doc": CompletionItemKind.File,
    "class": CompletionItemKind.Class,
    "envvar": CompletionItemKind.Variable,
    "function": CompletionItemKind.Function,
    "method": CompletionItemKind.Method,
    "module": CompletionItemKind.Module,
    "term": CompletionItemKind.Text,
}


class DomainHelpers:
    """Common methods that work on domains."""

    rst: SphinxLanguageServer

    @property
    def domains(self) -> Dict[str, Domain]:
        """Return a dictionary of known domains."""

        if self.rst.app is None or self.rst.app.env is None:
            return dict()

        return self.rst.app.env.domains  # type: ignore

    def get_default_domain(self, uri: str) -> Optional[str]:
        """Return the default domain for the given uri."""

        # TODO: Add support for .. default-domain::
        if self.rst.app is not None:
            return self.rst.app.config.primary_domain

        return None


class DomainDirectives(DirectiveLanguageFeature, DomainHelpers):
    """Support for directives coming from Sphinx's domains."""

    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst

        self._directives: Optional[Dict[str, Type[Directive]]] = None
        """Cache for known directives."""

    @property
    def directives(self) -> Dict[str, Type[Directive]]:
        if self._directives is not None:
            return self._directives

        directives = {}
        for prefix, domain in self.domains.items():
            for name, directive in domain.directives.items():
                directives[f"{prefix}:{name}"] = directive

        self._directives = directives
        return self._directives

    def get_implementation(
        self, directive: str, domain: Optional[str]
    ) -> Optional[Type[Directive]]:
        if domain is not None:
            return self.directives.get(f"{domain}:{directive}", None)

        if self.rst.app is None:
            return None

        # Try the default domain
        primary_domain = self.rst.app.config.primary_domain
        impl = self.directives.get(f"{primary_domain}:{directive}", None)
        if impl is not None:
            return impl

        # Try the std domain
        return self.directives.get(f"std:{directive}", None)

    def index_directives(self) -> Dict[str, Type[Directive]]:
        return self.directives

    def suggest_directives(
        self, context: CompletionContext
    ) -> Iterable[Tuple[str, Type[Directive]]]:
        # In addition to providing each directive fully qualified, we should provide a
        # suggestion for directives in the std and primary domains without the prefix.
        items = self.directives.copy()
        primary_domain = self.get_default_domain(context.doc.uri)

        for key, directive in self.directives.items():
            if key.startswith("std:"):
                items[key.replace("std:", "")] = directive
                continue

            if primary_domain and key.startswith(f"{primary_domain}:"):
                items[key.replace(f"{primary_domain}:", "")] = directive

        return items.items()

    def suggest_options(
        self, context: CompletionContext, directive: str, domain: Optional[str]
    ) -> Iterable[str]:
        impl = self.get_implementation(directive, domain)
        if impl is None:
            return []

        option_spec = getattr(impl, "option_spec", {})
        return option_spec.keys()


class DomainRoles(RoleLanguageFeature, DomainHelpers):
    """Support for roles coming from Sphinx's domains."""

    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst

        self._roles: Optional[Dict[str, Any]] = None
        """Cache for known roles."""

        self._role_target_types: Optional[Dict[str, List[str]]] = None
        """Cache for role target types."""

    @property
    def roles(self) -> Dict[str, Any]:
        if self._roles is not None:
            return self._roles

        roles = {}
        for prefix, domain in self.domains.items():
            for name, role in domain.roles.items():
                roles[f"{prefix}:{name}"] = role

        self._roles = roles
        return self._roles

    @property
    def role_target_types(self) -> Dict[str, List[str]]:
        if self._role_target_types is not None:
            return self._role_target_types

        self._role_target_types = {}

        for prefix, domain in self.domains.items():
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for name, item_type in domain.object_types.items():
                for role in item_type.roles:
                    role_key = fmt.format(name=role, prefix=prefix)
                    target_types = self._role_target_types.get(role_key, list())
                    target_types.append(name)

                    self._role_target_types[role_key] = target_types

        return self._role_target_types

    def _get_role_target_types(self, name: str, domain: str = "") -> List[str]:
        """Return a list indicating which object types a role is capable of linking
        with.
        """

        if domain:
            return self.role_target_types.get(f"{domain}:{name}", [])

        # Try the primary domain
        if self.rst.app and self.rst.app.config.primary_domain:
            key = f"{self.rst.app.config.primary_domain}:{name}"

            if key in self.role_target_types:
                return self.role_target_types[key]

        # Finally try the standard domain
        return self.role_target_types.get(f"std:{name}", [])

    def _get_role_targets(self, name: str, domain: str = "") -> List[tuple]:
        """Return a list of objects targeted by the given role.

        Parameters
        ----------
        name:
           The name of the role
        domain:
           The domain the role is a part of, if applicable.
        """

        targets: List[tuple] = []
        domain_obj: Optional[Domain] = None

        if domain:
            domain_obj = self.domains.get(domain, None)
        else:
            std = self.domains.get("std", None)
            if std and name in std.roles:
                domain_obj = std

            elif self.rst.app and self.rst.app.config.primary_domain:
                domain_obj = self.domains.get(self.rst.app.config.primary_domain, None)

        target_types = set(self._get_role_target_types(name, domain))

        if not domain_obj:
            self.rst.logger.debug(
                "Unable to find domain for role '%s:%s'", domain, name
            )
            return []

        for obj in domain_obj.get_objects():
            if obj[2] in target_types:
                targets.append(obj)

        return targets

    def get_implementation(
        self, role: str, domain: Optional[str]
    ) -> Optional[Directive]:
        if domain is not None:
            return self.roles.get(f"{domain}:{role}", None)

        if self.rst.app is None:
            return None

        # Try the default domain
        primary_domain = self.rst.app.config.primary_domain
        impl = self.roles.get(f"{primary_domain}:{role}", None)
        if impl is not None:
            return impl

        # Try the std domain
        return self.roles.get(f"std:{role}", None)

    def index_roles(self) -> Dict[str, Any]:
        return self.roles

    def suggest_roles(self, context: CompletionContext) -> Iterable[Tuple[str, Any]]:
        # In addition to providing each role fully qulaified, we should provide a
        # suggestion for directives in the std and primary domains without the prefix.
        items = self.roles.copy()
        primary_domain = self.get_default_domain(context.doc.uri)

        for key, role in self.roles.items():
            if key.startswith("std:"):
                items[key.replace("std:", "")] = role
                continue

            if primary_domain and key.startswith(f"{primary_domain}:"):
                items[key.replace(f"{primary_domain}:", "")] = role

        return items.items()

    def complete_targets(
        self, context: CompletionContext, name: str, domain: str
    ) -> List[CompletionItem]:
        label = context.match.group("label")

        # Intersphinx targets contain ':' characters.
        if ":" in label:
            return []

        return [
            object_to_completion_item(obj)
            for obj in self._get_role_targets(name, domain)
        ]

    def resolve_target_link(
        self, context: DocumentLinkContext, name: str, domain: Optional[str], label: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """``textDocument/documentLink`` support"""

        # Ignore intersphinx references.
        if ":" in label:
            return None, None

        # Other roles like :ref: do not make sense as the ``textDocument/documentLink``
        # api doesn't support specific locations like goto definition does.
        if (domain is not None and domain != "std") or name != "doc":
            return None, None

        return self.resolve_doc(context.doc, label), None

    def find_target_definitions(
        self, context: DefinitionContext, name: str, domain: str, label: str
    ) -> List[Location]:
        if not domain and name == "ref":
            return self.ref_definition(label)

        if not domain and name == "doc":
            return self.doc_definition(context.doc, label)

        return []

    def resolve_doc(self, doc: Document, label: str) -> Optional[str]:
        if self.rst.app is None:
            return None

        srcdir = self.rst.app.srcdir
        currentdir = pathlib.Path(Uri.to_fs_path(doc.uri)).parent

        if label.startswith("/"):
            path = pathlib.Path(srcdir, label[1:] + ".rst")
        else:
            path = pathlib.Path(currentdir, label + ".rst")

        if not path.exists():
            return None

        return Uri.from_fs_path(str(path))

    def doc_definition(self, doc: Document, label: str) -> List[Location]:
        """Goto definition implementation for ``:doc:`` targets"""

        uri = self.resolve_doc(doc, label)
        if not uri:
            return []

        return [
            Location(
                uri=uri,
                range=Range(
                    start=Position(line=0, character=0),
                    end=Position(line=1, character=0),
                ),
            )
        ]

    def ref_definition(self, label: str) -> List[Location]:
        """Goto definition implementation for ``:ref:`` targets"""

        if not self.rst.app or not self.rst.app.env:
            return []

        types = set(self._get_role_target_types("ref"))
        std = self.domains["std"]
        if std is None:
            return []

        docname = self.find_docname_for_label(label, std, types)
        if docname is None:
            return []

        path = self.rst.app.env.doc2path(docname)
        uri = Uri.from_fs_path(path)

        doctree = self.rst.get_initial_doctree(uri)
        if doctree is None:
            return []

        uri = None
        line = None

        for node in doctree.traverse(condition=nodes.target):
            if "refid" not in node:
                continue

            if doctree.nameids.get(label, "") == node["refid"]:
                uri = Uri.from_fs_path(node.source)
                line = node.line
                break

        if uri is None or line is None:
            return []

        return [
            Location(
                uri=uri,
                range=Range(
                    start=Position(line=line - 1, character=0),
                    end=Position(line=line, character=0),
                ),
            )
        ]

    def find_docname_for_label(
        self, label: str, domain: Domain, types: Optional[Set[str]] = None
    ) -> Optional[str]:
        """Given the label name and domain it belongs to, return the docname its
        definition resides in.

        Parameters
        ----------
        label
           The label to search for

        domain
           The domain to search within

        types
           A collection of object types that the label chould have.
        """

        docname = None
        types = types or set()

        # _, title, _, _, anchor, priority
        for name, _, type_, doc, _, _ in domain.get_objects():
            if types and type_ not in types:
                continue

            if name == label:
                docname = doc
                break

        return docname


class Intersphinx(RoleLanguageFeature):
    def __init__(self, rst: SphinxLanguageServer, domain: DomainRoles):
        self.rst = rst
        self.domain = domain

    def complete_targets(
        self, context: CompletionContext, name: str, domain: str
    ) -> List[CompletionItem]:
        label = context.match.group("label")

        # Intersphinx targets contain ':' characters.
        if ":" in label:
            return self.complete_intersphinx_targets(name, domain, label)

        return self.complete_intersphinx_projects(name, domain)

    def complete_intersphinx_projects(
        self, name: str, domain: str
    ) -> List[CompletionItem]:
        items = []
        for project in self.get_intersphinx_projects():
            if not self.has_intersphinx_targets(project, name, domain):
                continue

            items.append(
                CompletionItem(
                    label=project, detail="intersphinx", kind=CompletionItemKind.Module
                )
            )

        return items

    def complete_intersphinx_targets(
        self, name: str, domain: str, label: str
    ) -> List[CompletionItem]:
        items = []
        project, *_ = label.split(":")
        intersphinx_targets = self.get_intersphinx_targets(project, name, domain)

        for type_, targets in intersphinx_targets.items():
            items += [
                intersphinx_target_to_completion_item(project, label, target, type_)
                for label, target in targets.items()
            ]

        return items

    def resolve_target_link(
        self, context: DocumentLinkContext, name: str, domain: Optional[str], label: str
    ) -> Tuple[Optional[str], Optional[str]]:
        if not self.rst.app:
            return None, None

        project, *parts = label.split(":")
        label = ":".join(parts)
        targets = self.get_intersphinx_targets(project, name, domain or "")

        for _, items in targets.items():
            if label in items:
                source, version, url, display = items[label]
                name = label if display == "-" else display
                tooltip = f"{name} - {source} v{version}"

                return url, tooltip

        return None, None

    def get_intersphinx_projects(self) -> List[str]:
        """Return the list of configured intersphinx project names."""

        if self.rst.app is None:
            return []

        inv = getattr(self.rst.app.env, "intersphinx_named_inventory", {})
        return list(inv.keys())

    def has_intersphinx_targets(
        self, project: str, name: str, domain: str = ""
    ) -> bool:
        """Return ``True`` if the given intersphinx project has targets targeted by the
        given role.

        Parameters
        ----------
        project
           The project to check

        name
           The name of the role

        domain
           The domain the role is a part of, if applicable.
        """
        targets = self.get_intersphinx_targets(project, name, domain)

        if len(targets) == 0:
            return False

        return any([len(items) > 0 for items in targets.values()])

    def get_intersphinx_targets(
        self, project: str, name: str, domain: str = ""
    ) -> Dict[str, Dict[str, tuple]]:
        """Return the intersphinx objects targeted by the given role.

        Parameters
        ----------
        project
           The project to return targets from

        name
           The name of the role

        domain
           The domain the role is a part of, if applicable.
        """

        if self.rst.app is None:
            return {}

        inv = getattr(self.rst.app.env, "intersphinx_named_inventory", {})
        if project not in inv:
            return {}

        targets = {}
        inv = inv[project]

        for target_type in self.domain._get_role_target_types(name, domain):
            explicit_domain = f"{domain}:{target_type}"
            if explicit_domain in inv:
                targets[target_type] = inv[explicit_domain]
                continue

            primary_domain = f'{self.rst.app.config.primary_domain or ""}:{target_type}'
            if primary_domain in inv:
                targets[target_type] = inv[primary_domain]
                continue

            std_domain = f"std:{target_type}"
            if std_domain in inv:
                targets[target_type] = inv[std_domain]

        return targets


def intersphinx_target_to_completion_item(
    project: str, label: str, target: tuple, type_: str
) -> CompletionItem:
    # _. _. url, _
    source, version, _, display = target

    display_name = label if display == "-" else display
    completion_kind = ":".join(type_.split(":")[1:]) if ":" in type_ else type_

    if version:
        version = f" v{version}"

    return CompletionItem(
        label=label,
        detail=f"{display_name} - {source}{version}",
        kind=TARGET_KINDS.get(completion_kind, CompletionItemKind.Reference),
        insert_text=f"{project}:{label}",
    )


def object_to_completion_item(object_: tuple) -> CompletionItem:
    # _, _, _, docname, anchor, priority
    name, display_name, type_, _, _, _ = object_
    insert_text = name

    key = type_.split(":")[1] if ":" in type_ else type_
    kind = TARGET_KINDS.get(key, CompletionItemKind.Reference)

    # ensure :doc: targets are inserted as an absolute path - that way the reference
    # will always work regardless of the file's location.
    if type_ == "doc":
        insert_text = f"/{name}"

    # :option: targets need to be inserted as `<progname> <option>` in order to resolve
    # correctly. However, this only seems to be the case "locally" as
    # `<progname>.<option>` seems to resolve fine when using intersphinx...
    if type_ == "cmdoption":
        name = " ".join(name.split("."))
        display_name = name
        insert_text = name

    return CompletionItem(
        label=name, kind=kind, detail=str(display_name), insert_text=insert_text
    )


def esbonio_setup(rst: SphinxLanguageServer, directives: Directives, roles: Roles):
    directives.add_feature(DomainDirectives(rst))

    domain_roles = DomainRoles(rst)
    intersphinx = Intersphinx(rst, domain_roles)

    roles.add_feature(domain_roles)
    roles.add_feature(intersphinx)
